#ifndef GOAPI_LOGGING_H
#define GOAPI_LOGGING_H

#include <GoApi/GoApiDef.h>

namespace Go
{
    class Logging 
    {
    public:
        Logging() {};

        static GoApiCppFx(kStatus) Go_Debug_Log(int line, 
                                             const kChar* fileName, 
                                             kLogOption options, 
                                             const kChar* source, 
                                             const kChar* format, ...);

        static GoApiCppFx(kStatus) FormatLogBuffer(kDateTime dateTime, kLogOption options, const kChar* source, 
                                                const kChar* message, kChar* buffer, kSize capacity);

        static GoApiCppFx(kStatus) StripPositionalParameters(const kChar* orig, kChar* stripped, kSize capacity);

        /**
         * Strips trailing characters from a given string.
         *
         * @param    str                 String to remove trailing characters from.
         * @param    stripSet            The set of trailing characters to remove.
         *
         *  @code {.cpp}
         *  std::string trimMe = "Trim me...,,,";
         *  Go::Logging::StripTrailing(trimeMe, ".,");
         *  ASSERT_STREQ(trimMe.c_str(), "Trim me");
         *  @endcode
         */
        static GoApiCppFx(std::string) StripTrailing(std::string str, std::string stripSet);

    private:
        static GoApiCppFx(kStatus) FormatLogOption(kLogOption options, kChar* buffer, kSize bufferSize);
    };
}

// See (FSS-1301) and kApiDef.h:
// The lower 32 bits of an option value are reserved for platform-defined 
//  purposes; applications can use the upper 32 bits as desired. 

// Individual Log option bit masks.
#define GO_LOG_OPTION_PLAT_MASK                     (0x0000000000000010)

#define GO_LOG_OPTION_GOCAPP_MASK                   (0x8000000000000000)
#define GO_LOG_OPTION_GOCAPP_USER_MASK              (0x0000000100000000)

#define GO_LOG_OPTION_LEVEL_MASK                    (0x0000000000000003)
#define GO_LOG_OPTION_LEVEL_INFO                    (0x0)
#define GO_LOG_OPTION_LEVEL_WARN                    (0x1)
#define GO_LOG_OPTION_LEVEL_ERROR                   (0x2)

// Commonly used options (we hard code them instead of bit shifting them each time.
#define GO_LOG_OPTION_GOCAPP_INFO                   ((kLogOption) 0x8000000000000000)  // (GO_LOG_OPTION_GOC_APP_MASK | GO_LOG_OPTION_LEVEL_INFO)
#define GO_LOG_OPTION_GOCAPP_WARN                   ((kLogOption) 0x8000000000000001)  // (GO_LOG_OPTION_GOC_APP_MASK | GO_LOG_OPTION_LEVEL_WARN)
#define GO_LOG_OPTION_GOCAPP_ERROR                  ((kLogOption) 0x8000000000000002)  // (GO_LOG_OPTION_GOC_APP_MASK | GO_LOG_OPTION_LEVEL_ERROR)
#define GO_LOG_OPTION_GOCAPP_USER_INFO              ((kLogOption) 0x8000000100000000)  // (GO_LOG_OPTION_GOCAPP_INFO  | GO_LOG_OPTION_GOCAPP_USER_MASK | GO_LOG_OPTION_GOCAPP_TOOL_MASK)
#define GO_LOG_OPTION_GOCAPP_USER_WARN              ((kLogOption) 0x8000000100000001)  // (GO_LOG_OPTION_GOCAPP_WARN  | GO_LOG_OPTION_GOCAPP_USER_MASK | GO_LOG_OPTION_GOCAPP_TOOL_MASK)
#define GO_LOG_OPTION_GOCAPP_USER_ERROR             ((kLogOption) 0x8000000100000002)  // (GO_LOG_OPTION_GOCAPP_ERROR | GO_LOG_OPTION_GOCAPP_USER_MASK | GO_LOG_OPTION_GOCAPP_TOOL_MASK)

// Used to check what options are set.
#define GoLogIsPlatform(option)                     (((option) & GO_LOG_OPTION_PLAT_MASK) > 0)
#define GoLogIsGocApplication(option)               (((option) & GO_LOG_OPTION_GOCAPP_MASK) > 0)
#define GoLogIsGocUser(option)                      (((option) & GO_LOG_OPTION_GOCAPP_USER_MASK) > 0)       // User-facing system messages.

#define GoLogIsInfo(option)                         (((option) & GO_LOG_OPTION_LEVEL_MASK) == GO_LOG_OPTION_LEVEL_INFO)
#define GoLogIsWarn(option)                         (((option) & GO_LOG_OPTION_LEVEL_MASK) == GO_LOG_OPTION_LEVEL_WARN)
#define GoLogIsError(option)                        (((option) & GO_LOG_OPTION_LEVEL_MASK) == GO_LOG_OPTION_LEVEL_ERROR)

// General purpose logging macros.  Use this to log items that can be hot-clickable
// when running within Visual Studio, 
// NOTE: Child Process Debugging must be enabled for Google tests.


/**
 * Helper macro to log an informational message.
 *
 * @param   format      Informational message to be logged.
 */
#define GoLogInfo(format, ...)                      Go::Logging::Go_Debug_Log(__LINE__, __FILE__, GO_LOG_OPTION_GOCAPP_INFO, __FUNCTION__, format, ## __VA_ARGS__)

 /**
 * Helper macro to log a warning message.
 *
 * @param   format      Warning message to be logged.
 */
#define GoLogWarn(format, ...)                      Go::Logging::Go_Debug_Log(__LINE__, __FILE__, GO_LOG_OPTION_GOCAPP_WARN, __FUNCTION__, format, ## __VA_ARGS__)

 /**
 * Helper macro to log an error message.
 *
 * @param   format      Error message to be logged.
 */
#define GoLogError(format, ...)                     Go::Logging::Go_Debug_Log(__LINE__, __FILE__, GO_LOG_OPTION_GOCAPP_ERROR, __FUNCTION__, format, ## __VA_ARGS__)

/**
 * Helper macro to log a user informational message.
 * NOTE: The message should be marked for localization with pgettext() macro.
 *
 * @param   source      Source of the message.
 * @param   format      User informational message to be logged.
 */
#define GoLogUserInfo(source, format, ...)          Go::Logging::Go_Debug_Log(__LINE__, __FILE__, GO_LOG_OPTION_GOCAPP_USER_INFO, source, format, ## __VA_ARGS__)

/**
 * Helper macro to log a user warning message.
 * NOTE: The message should be marked for localization with pgettext() macro.
 *
 * @param   source      Source of the message.
 * @param   format      User warning message to be logged.
 */
#define GoLogUserWarn(source, format, ...)          Go::Logging::Go_Debug_Log(__LINE__, __FILE__, GO_LOG_OPTION_GOCAPP_USER_WARN, source, format, ## __VA_ARGS__)

/**
 * Helper macro to log a user error message.
 * NOTE: The message should be marked for localization with pgettext() macro.
 *
 * @param   source      Source of the message.
 * @param   format      User error message to be logged.
 */
#define GoLogUserError(source, format, ...)         Go::Logging::Go_Debug_Log(__LINE__, __FILE__, GO_LOG_OPTION_GOCAPP_USER_ERROR, source, format, ## __VA_ARGS__)

// GOS-5108: Added GoLogDebug macros which act as GoLogUser in Debug builds and GoLog in Release builds.
// First set of macros are used in Debug builds.
#ifdef _DEBUG

/**
 * Helper macro to log a user info message in debug builds and a normal info message in release builds.
 *
 * @param   source      Source of the message.
 * @param   format      User info message to be logged.
 */
#define GoLogDebugInfo(source, format, ...)         GoLogUserInfo(source, format, ## __VA_ARGS__)

/**
 * Helper macro to log a user warning message in debug builds and a normal warning message in release builds.
 *
 * @param   source      Source of the message.
 * @param   format      User warning message to be logged.
 */
#define GoLogDebugWarn(source, format, ...)         GoLogUserWarn(source, format, ## __VA_ARGS__)

/**
 * Helper macro to log a user error message in debug builds and a normal error message in release builds.
 *
 * @param   source      Source of the message.
 * @param   format      User error message to be logged.
 */
#define GoLogDebugError(source, format, ...)        GoLogUserError(source, format, ## __VA_ARGS__)

// GOS-5108: Release builds use this second set of macros.
#else

/**
 * Helper macro to log a user info message in debug builds and a normal info message in release builds.
 *
 * @param   source      Source of the message.
 * @param   format      User info message to be logged.
 */
#define GoLogDebugInfo(source, format, ...)         GoLogInfo(format, ## __VA_ARGS__)

/**
 * Helper macro to log a user warning message in debug builds and a normal warning message in release builds.
 *
 * @param   source      Source of the message.
 * @param   format      User warning message to be logged.
 */
#define GoLogDebugWarn(source, format, ...)         GoLogWarn(format, ## __VA_ARGS__)

/**
 * Helper macro to log a user error message in debug builds and a normal error message in release builds.
 *
 * @param   source      Source of the message.
 * @param   format      User error message to be logged.
 */
#define GoLogDebugError(source, format, ...)        GoLogError(format, ## __VA_ARGS__)

#endif


#endif