/**
 * @file    GoTransaction.h
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
#ifndef GO_PXL_SDK_TRANSACTION_H
#define GO_PXL_SDK_TRANSACTION_H

#include <GoPxLSdk/GoRequest.h>
#include <GoPxLSdk/GoResponse.h>

namespace GoPxLSdk
{
constexpr k64u DEFAULT_TRANSACTION_TIMEOUT_USEC = 3 * 1000 * 1000;  // 3,000,000 microseconds = 3 seconds.

class GoPxLSdkClass GoTransaction
{
public:
    /**
     * Constructs GoTransaction.
     *
     * @public                @memberof GoTransaction
     * @version               Introduced in 0.2.1.53
     */
    GoTransaction() = default;

    /**
     * Transaction handle to simplify working with asynchronous result data.
     *
     * @public                @memberof GoTransaction
     * @version               Introduced in 0.2.1.53
     * @param request Request to handle
     */
    GoTransaction(GoRequest request);

    /**
     * Checks the future transaction response value without returning it. Waits until the response is available.
     *
     * @public                @memberof GoTransaction
     * @version               Introduced in 0.2.1.53
     * @param timeout         Time to wait for the response in microseconds. Default value is `TRANSACTION_TIMEOUT`.
     * @throws GoRequestError If the server responds with a fail status for the request.
     * @throws GoChannelError If the request failed for unexpected, channel related reasons such as timeout.
     */
    void CheckResponse(k64u timeout = DEFAULT_TRANSACTION_TIMEOUT_USEC);

    /**
     * Gets the future transaction response value. Waits until the response is available.
     *
     * @public                @memberof GoTransaction
     * @version               Introduced in 0.2.1.53
     * @param timeout         Time to wait for the response in microseconds. Default value is `TRANSACTION_TIMEOUT`.
     * @return                The request's response.
     * @throws GoRequestError If the server responds with a fail status for the request.
     * @throws GoChannelError If the request failed for unexpected, channel related reasons such as timeout.
     */
    GoRequestResponse GetResponse(k64u timeout = DEFAULT_TRANSACTION_TIMEOUT_USEC);

    /**
     * Gets the transaction request value. Since a transaction is initiated with a request, this method
     * should always succeed.
     *
     * @public                @memberof GoTransaction
     * @version               Introduced in 0.2.1.53
     */
    GoRequest GetRequest();

    /**
     * Gets the future which stores the transaction promise.
     *
     * @public                @memberof GoTransaction
     * @version               Introduced in 0.2.1.53
     * @return The Transaction Promise which holds the response.
     */
    void OnResponse(const std::shared_ptr<GoRequestResponse>& response);

    /**
     * Gets the future which stores the transaction response value.
     *
     * @public                @memberof GoTransaction
     * @version               Introduced in 0.2.1.53
     * @return The Future which holds the response.
     */
    std::shared_future<GoRequestResponse>& GetResponseFuture();

private:
    GoRequest request;
    std::shared_ptr<std::promise<GoRequestResponse>> promise;
    std::shared_future<GoRequestResponse> futureResponse;
};

}

#endif
