/*
 * @file            ConfigureSensor.cpp
 * Objective        Connect to a Gocator sensor system to change sensor configurations.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

#include <kApi/kApiDef.h>
#include <GoApi/GoApiLib.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoRequestError.h>
#include <GoPxLSdk/GoChannelError.h>

using std::string;
using namespace GoPxLSdk;

#define SENSOR_IP "192.168.1.10"

const string SENSOR_SERIAL_NUMBER = "95246";
const string ENGINE_ID = "g2xxx";
const string SCANNER_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0";
const string SENSOR_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/sensors/sensor-0";
const string IPCONFIG_PATH = "/environ/ipConfig";
const string VISIBLE_SENSORS_PATH = "/scan/visibleSensors/";
const int ERROR_STATUS = -1;
const int OK_STATUS = 0;
const int REST_COMMAND_TIMEOUT = 3'000'000; // 3 seconds.

int ConfigureSensor()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);

    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));

    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
                                 "also check if you are using the correct IP address and port number."
                  << std::endl;
        return ERROR_STATUS;
    }

    // Configuration examples
    std::cout << "\nConfiguring sensor with the serial number of " << SENSOR_SERIAL_NUMBER << "." << std::endl;

    // Configuration 1: Update trigger source.
    const std::map<int, string> triggerSources{{0, "Time"}, {1, "Encoder"}, {2, "External Input"}, {3, "Software"}};
    GoJson payload = GoJson("{\"parameters\" : {\"triggerSettings\" : {\"source\" : 0} }}");
    try
    {
        system.Client().Update(SCANNER_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        response = system.Client().Read(SCANNER_PATH).GetResponse().Payload();
        auto new_value = triggerSources.at(response.At("/parameters/triggerSettings/source").Get<int>());
        std::cout << "\nTrigger source set to \"" << new_value << "\"." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Check if the API path is valid." << std::endl;
        return ERROR_STATUS;
    }

    // Configuration 2: Update exposure.
    payload = GoJson("{\"parameters\" : {\"exposureSettings\" : {\"exposureMode\" : 0, \"singleExposure\" : 400} }}");

    try
    {
        system.Client().Update(SENSOR_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        response = system.Client().Read(SENSOR_PATH).GetResponse().Payload();
        auto new_value = response.At("/parameters/exposureSettings/singleExposure").Get<int>();
        std::cout << "\nExposure set to \"" << new_value << "\"." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Check if the API path is valid." << std::endl;
        return ERROR_STATUS;
    }

    // Configuration 3: Update multiple exposures.
    payload = GoJson("{\"parameters\" : {\"exposureSettings\" : {\"exposureMode\" : 1, \"multipleExposures\" : [400.0, 600.0, 800.0]} }}");
    system.Client().Update(SENSOR_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);

    response = system.Client().Read(SENSOR_PATH).GetResponse().Payload();
    std::cout << "\nMulti-exposure set to " << response.At("/parameters/exposureSettings/multipleExposures") << "." << std::endl;

    // Configuration 4: Set active area.
    payload = GoJson("{\"parameters\" : {\"activeAreaSettings\" : {\"activeArea\" : {\"x\": 0.2 }}}}");
    system.Client().Update(SENSOR_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);

    response = system.Client().Read(SENSOR_PATH).GetResponse().Payload();
    std::cout << "\nActive Area updated to " << response.At("/parameters/activeAreaSettings/activeArea") << "." << std::endl;

    // Configuration 5: Update network configuration.
    response = system.Client().Read(IPCONFIG_PATH).GetResponse().Payload();
    std::cout << "\nNetwork configurations: " << response.At("/interfaces") << "." << std::endl;

    // Uncomment to update the current IP address to 192.168.1.11.
    /*
    response.Replace<string>("/interfaces/0/ipAddress", "192.168.1.11");
    system.Client().Update(IPCONFIG_PATH, response).CheckResponse(REST_COMMAND_TIMEOUT);
    std::cout << "Updated IP address of the 1st interface to " << response.At("/interfaces/0/ipAddress") << "." << std::endl;
    */

    // Disconnect the sensor system.
    system.Disconnect();

    return OK_STATUS;
}

int main(int argc, char **argv)
{
    int status;

    try
    {
        status = ConfigureSensor();
    }
    catch (const GoRequestError &e)
    {
        string errorMessage = "";

        const GoJson &payload = e.GetResponse().Payload();
        if (!payload.Empty() && payload.HasKey("errors"))
        {
            GoJson errors = payload.At("errors");
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }

        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError &e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    std::cout << "\nPress ENTER to exit the program..." << std::endl;
    getchar();

    return status;
}
