/*
 * @file            Discover.cpp
 * Objective        Discover sensor and GoPxL instances on the network.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 *
 * Objective: Discover sensor and GoPxL instances on the network.
 */

#include <kApi/kApiDef.h>
#include <GoApi/GoApiLib.h>
#include <GoPxLSdk/GoDiscoveryClient.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>

using std::string;
using namespace GoPxLSdk;

const string ENV_CONFIG_PATH = "/environ/ipConfig";
const string VISIBLE_SENSORS_PATH = "/scan/visibleSensors/";
const int ERROR_STATUS = -1;
const int OK_STATUS = 0;
const int DISCOVER_TIMEOUT = 3'000'000; // 3 seconds.

int DiscoverSensors()
{
    kStatus status;
    kAssembly goApiLib = kNULL;

    GoJson response;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Discover sensors.
    auto discovery = std::make_unique<GoDiscoveryClient>();
    discovery->BlockingDiscover(DISCOVER_TIMEOUT);
    auto& instances = discovery->InstanceList();
    std::cout << "Number of sensors on the network: " << instances.size() << std::endl;

    if (instances.empty())
    {
        std::cout << "No sensors found, make sure sensors or GoPxl on PC/GoMax are available and connected." << std::endl;
        return ERROR_STATUS;
    }

    // Iterate over the sensors to get their information.
    int instanceCount = 0;
    for (size_t i = 0; i < instances.size(); i++)
    {
        kChar ipAddress[16], gateway[32], mask[32];
        const GoInstance& instance = discovery->InstanceList().at(i);

        kIpAddress_Format(instance.GetIpAddress(), ipAddress, sizeof(ipAddress));
        kIpAddress_Format(instance.GetGateway(), gateway, sizeof(gateway));
        kIpAddress_Format(instance.GetMask(), mask, sizeof(mask));

        std::cout << std::endl;
        std::cout << string(52, '*') << " GoPxL instance " << instanceCount + 1 << std::endl; // Prints 52 asterisks to help the readability.
        std::cout << "IP Address: " << ipAddress << std::endl;
        std::cout << "Gateway: " << string(gateway) << std::endl;
        std::cout << "Mask: " << mask << std::endl;
        std::cout << "DHCP: " << (instance.GetIsDhcp() ? "True" : "False") << std::endl;
        std::cout << "Application ID: " << instance.GetAppId() << std::endl;
        std::cout << "Application Name: " << instance.GetAppName() << std::endl;
        std::cout << "Application Version: " << instance.GetAppVersion() << std::endl;
        std::cout << "Control Port: " << instance.GetControlPort() << std::endl;
        std::cout << "GDP Port: " << instance.GetGdpPort() << std::endl;
        std::cout << "Web Port: " << instance.GetWebPort() << std::endl;
        std::cout << "IP Conflict: " << (instance.GetIsAddressConflict() ? "Yes" : "No") << std::endl;
        ++instanceCount;

        // Connect to a sensor/instance.
        auto system = GoSystem(instance.GetIpAddress(), instance.GetControlPort());
        try
        {
            system.Connect();
        }
        catch (const GoRequestError& e)
        {
            std::cerr << e.what() << " - Connection failed." << std::endl;
        }
        
        try
        {
            // If this is a multi-sensor system, iterate to get each sensor information.
            GoJson sensors = system.Client().Read(VISIBLE_SENSORS_PATH).GetResponse().Payload().At("/sensors");
            int sensorCount = 0;
            for (auto sensor = sensors.Begin(); sensor != sensors.End(); sensor.operator++(0))
            {
                auto sensorPath = system.SensorPath(sensor.Value().Get<string>("serialNumber"));
                if (!sensorPath.empty()) {
                    auto sensorSerialNumber = sensor.Value().At("/serialNumber").Get<string>();
                    GoJson ipConfig = system.Client().Read(ENV_CONFIG_PATH).GetResponse().Payload().At("/interfaces");

                    std::cout << "\n" << string(22, '*') << " Sensor " << sensorCount + 1 << std::endl;
                    std::cout << "Local sensor: " << (sensor.Value().At("/isLocal").Get<bool>() ? "Yes" : "No") << std::endl;
                    std::cout << "IP Address: " << sensor.Value().At("/parameters/ipAddress").Get<string>() << std::endl;
                    std::cout << "Gateway Address: " << ipConfig.At("/0/gatewayAddress").Get<string>() << std::endl;
                    std::cout << "Subnet Mask: " << ipConfig.At("/0/subnetMask").Get<string>() << std::endl;
                    std::cout << "DHCP: " << ipConfig.At("/0/dhcp") << std::endl;
                    std::cout << "Serial Number: " << sensorSerialNumber << std::endl;
                    std::cout << "Model: " << sensor.Value().At("/model").Get<string>() << std::endl;
                    std::cout << "Engine ID: " << sensor.Value().At("/engineId").Get<string>() << std::endl;
                    std::cout << "Sensor Path: " << sensorPath << std::endl;
                }
                else
                {
                    continue;
                }
                ++sensorCount;
            }
        }
        catch (const GoRequestError& e)
        {
            std::cout << e.what() << " - Failed to connect. Sensor might be controlled by another instance." << std::endl;
        }
    }

    return OK_STATUS;
}

int main(int argc, char** argv)
{
    int status;
    try
    {
        status = DiscoverSensors();
    }
    catch (const GoRequestError& e)
    {
        string errorMessage = "";

        const GoJson& payload = e.GetResponse().Payload();
        if (!payload.Empty() && payload.HasKey("errors"))
        {
            GoJson errors = payload.At("errors");
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }
        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError& e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    std::cout << "\nPress ENTER to exit the program..." << std::endl;
    getchar();

    return status;
}
