/*
 * @file            ReceiveSurface.cpp
 * Objective        Connect to the GoPxL Data Protocol and receive surface data.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 *
 * GoGdpClient::ReceiveDataSync() or GoGdpClient::ReceiveDataAsync() can be used to receive data from the GoPxL Data Protocol (GDP) serve:
    1. Pass a timeout value (in microseconds) within which data will be received from the server.
    2. The GDP server sends out individual data messages. The last message is marked with an isLastMsg flag.
    3. ReceiveDataSync() receives individual data messages and reassembles to a data set (GoDataSet).
    4. GoGdpClient::DataSet() can be used to access the data set.
    5. GoGdpMsg::Type() indicates the type of individual messages in the dataset as described in GoPxLSdk::MessageType.
 */

#include <GoApi/GoApiLib.h>
#include <kApi/kApiDef.h>
#include <kApi/Data/kArray2.h>
#include <GoPxLSdk/GoSystem.h>
#include <GoPxLSdk/GoChannelError.h>
#include <GoPxLSdk/GoRequestError.h>
#include <GoPxLSdk/GoGdpClient.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpStamp.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpSurfaceUniform.h>
#include <GoPxLSdk/GoGdpMsg/GoGdpSurfacePointCloud.h>

using std::string;
using namespace GoPxLSdk;

#define SENSOR_IP "192.168.1.10"

const string ENGINE_ID = "g2xxx";
const string OUTPUTS_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/outputs";
const string UNIFORM_SURFACE_OUTPUT_PATH = OUTPUTS_PATH + "/topUniformSurface";
const string POINT_CLOUD_SURFACE_OUTPUT_PATH = OUTPUTS_PATH + "/sensor-0PointcloudSurface";
const string GOCATOR_CONTROL_PATH = "/controls/gocator";
const string GOCATOR_ADD_OUTPUT_PATH = "/controls/gocator/outputs/commands/add";
const int ERROR_STATUS = -1;
const int OK_STATUS = 0;
const int REST_COMMAND_TIMEOUT = 3'000'000; // Duration in microseconds.
const int RECEIVE_DATA_TIMEOUT = 3'000'000; // Duration in microseconds.

typedef struct SurfacePoint
{
    double x; // x coordinate - position along laser line (mm).
    double y; // y coordinate - position along the direction of travel (mm).
    double z; // z coordinate - height at a given x position(mm).
    unsigned char intensity;
} SurfacePoint;

typedef struct UniformSurface
{
    /* The x and y coordinates of a given point in the height map can be calculated as following:
     * for z value of HeightMap[i][j]
     * double x = xOffset + xResolution * i;
     * double y = yOffset + yResolution * j;
     */

    UniformSurface(int length, int width, double XOffset, double YOffset, double XResolution, double YResolution)
    {
        HeightMap.resize(length);
        for (int i = 0; i < length; i++)
        {
            HeightMap[i] = std::vector<double>(width);
        }
        xOffset = XOffset;
        yOffset = YOffset;
        xResolution = XResolution;
        yResolution = YResolution;
    }

    std::vector<std::vector<double>> HeightMap; // Z coordinates - height at a given x and y position (mm).
    double xOffset;                             // x Offset - offset position along laser line (mm).
    double yOffset;                             // y Offset - offset position along the direction of travel (mm).
    double xResolution;                         // x Resolution - resolution along laser line (mm).
    double yResolution;                         // y resolution - resolution along the direction of travel(mm).

} UniformSurface;

int ReceiveSurface()
{
    GoSystem system;
    kStatus status;
    kAssembly goApiLib = kNULL;
    GoJson response;
    GoJson payload;

    std::vector<std::vector<SurfacePoint>> surfaceBuffer;

    // Constructs GoPxL API core framework.
    if ((status = GoApiLib_Construct(&goApiLib)) != kOK)
    {
        std::cout << "Error: " << status << " - Failed to construct GoApiLib." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to a sensor system.
    kIpAddress systemIpAddress;
    kIpAddress_Parse(&systemIpAddress, SENSOR_IP);
    system = GoSystem(systemIpAddress, GO_PXL_SDK_DEFAULT_CONTROL_PORT);

    kChar ipAddress[16];
    kIpAddress_Format(system.Address(), ipAddress, sizeof(ipAddress));

    try
    {
        system.Connect();
        std::cout << "\nConnected to " << ipAddress << ":" << system.ControlPort() << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Connection failed. Check if sensor is powered on and connected "
                                 "also check if you are using the correct IP address and port number."
                  << std::endl;
        return ERROR_STATUS;
    }

    // Run the sensor.
    if (system.RunningState() == GoSystem::State::Ready)
    {
        system.Start();
        std::cout << "\nSystem started." << std::endl;
    }

    // Enable Gocator Protocol.
    payload = GoJson("{\"enabled\": true}");
    try
    {
        system.Client().Update(GOCATOR_CONTROL_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "\nGocator Protocol enabled." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " Check the timeout value or the API path." << std::endl;
        return ERROR_STATUS;
    }

    // Add surface to Gocator Protocol output.
    try
    {
        payload = GoJson("{\"enabled\": true}");
        // response = system.Client().Read(POINT_CLOUD_SURFACE_OUTPUT_PATH).GetResponse().Payload();
        response = system.Client().Read(UNIFORM_SURFACE_OUTPUT_PATH).GetResponse().Payload();
        string surfaceDataSourceId = response.At("dataSourceId").ToString();
        std::cout << surfaceDataSourceId << std::endl;
        payload = GoJson("{\"source\":" + surfaceDataSourceId + ", \"outputId\":0, \"autoShift\":true}");
        system.Client().Call(GOCATOR_ADD_OUTPUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
        std::cout << "\nAdded surface to the Gocator Protocol output." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << e.what() << " - Failed to add surface data to Gocator Protocol output. Check if uniform/point cloud toggle is correctly set." << std::endl;
        return ERROR_STATUS;
    }

    // Connect to the Gocator Data Protocol (GDP) server.
    auto gdpClient = std::make_unique<GoGdpClient>();
    try
    {
        gdpClient->Connect(system.Address(), GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT);
        std::cout << "\nGocator Protocol connected." << std::endl;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << " - Failed to connect to the GDP server." << std::endl;
        return ERROR_STATUS;
    }

    // Establish a synchronous data transfer.
    bool surfaceReceived = false;
    time_t timeout = time(NULL) + 10; // Define a 10-second timeout
    std::cout << "\nWaiting until a dataset with Surface data is received within 10 seconds." << std::endl;

    // Repeat receiving data within the specified timeout until a Surface data is received.
    while (!surfaceReceived && time(NULL) < timeout)
    {
        // Establish a synchronous data transfer with a 3-second timeout.
        try
        {
            gdpClient->ReceiveDataSync(RECEIVE_DATA_TIMEOUT);
            std::cout << "\nSynchronous data transfer established." << std::endl;
        }
        catch (const std::exception& e)
        {
            std::cerr << e.what() << " - Failed to receive data via synchronous connection." << std::endl;
            return ERROR_STATUS;
        }

        std::cout << "."; // Output a dot indicating loop is still running.

        // Receive data.
        const GoDataSet &receivedDataSet = gdpClient->DataSet();

        if (receivedDataSet.Count() < 1)
        {
            std::cout << "No data is received." << std::endl;
            return ERROR_STATUS;
        }

        // Loop through messages in the dataSet.
        for (size_t i = 0; i < receivedDataSet.Count(); i++)
        {
            if (receivedDataSet.GdpMsgAt(i).Type() == GoPxLSdk::MessageType::UNIFORM_SURFACE ||
                receivedDataSet.GdpMsgAt(i).Type() == GoPxLSdk::MessageType::SURFACE_POINT_CLOUD)
            {
                std::cout << "\nA dataset with a Surface message has been received." << std::endl;
                surfaceReceived = true;
                break;
            }
        }
    }

    // Get the data set again so that it is available outside the while loop scope.
    const GoDataSet &receivedDataSet = gdpClient->DataSet();

    std::cout << "\nTotal number of messages in this data set: " << receivedDataSet.Count() << std::endl;
    // Loop through messages in the dataSet.
    for (size_t i = 0; i < receivedDataSet.Count(); i++)
    {
        std::cout << "\nMessage " << i + 1 << " of " << receivedDataSet.Count() << std::endl;
        std::cout << "Message Type " << (int)receivedDataSet.GdpMsgAt(i).Type() << ": " << receivedDataSet.GdpMsgAt(i).DataSourceId() << std::endl;

        switch (receivedDataSet.GdpMsgAt(i).Type())
        {

        case GoPxLSdk::MessageType::STAMP:
        {
            const GoGdpStamp &stampMsg = (const GoGdpStamp &)receivedDataSet.GdpMsgAt(i);
            std::cout << "\nStamp" << std::endl;
            std::cout << "Frame index: " << stampMsg.FrameIndex() << std::endl;
            std::cout << "Timestamp: " << stampMsg.Timestamp() << std::endl;
            std::cout << "Encoder: " << stampMsg.Encoder() << std::endl;
            std::cout << "Encoder at Z: " << stampMsg.EncoderAtZ() << std::endl;
            std::cout << "Status: " << stampMsg.Status() << std::endl;
            break;
        }

        case GoPxLSdk::MessageType::UNIFORM_SURFACE:
        {
            const GoGdpSurfaceUniform &uniformSurfaceMsg = (const GoGdpSurfaceUniform &)receivedDataSet.GdpMsgAt(i);
            std::cout << "\nUniform Surface" << std::endl;
            std::cout << "Date Source: " << uniformSurfaceMsg.DataSourceId() << std::endl;
            std::cout << "Array count: " << uniformSurfaceMsg.ArrayedCount() << std::endl;
            std::cout << "Array index: " << uniformSurfaceMsg.ArrayedIndex() << std::endl;

            // Transform to real world coordinates.
            auto length = uniformSurfaceMsg.Length();
            auto width = uniformSurfaceMsg.Width();
            auto intensityArray = uniformSurfaceMsg.Intensities();
            auto intensityLength = uniformSurfaceMsg.IntensityLength();
            auto intensityWidth = uniformSurfaceMsg.IntensityWidth();

            UniformSurface uniformSurface = UniformSurface(length, width,
                                                           uniformSurfaceMsg.Offset().x,
                                                           uniformSurfaceMsg.Offset().y,
                                                           uniformSurfaceMsg.Resolution().x,
                                                           uniformSurfaceMsg.Resolution().y);

            /*  The surface data is sent in a 2D array of z values.
             *  The x and y real-world coordinates can be calculated from the indicies of the 2D array,
             *  using the x and y offsets and resolutions.
             *  The z values must be transformed into real world coordinates before being used.
             */
            for (k32u j = 0; j < length; j++)
            {
                for (k32u k = 0; k < width; k++)
                {
                    k16s data;
                    kArray2_Item(uniformSurfaceMsg.Ranges(), j, k, &data);

                    if (data != k16S_NULL)
                    {
                        // Calculates the Z position in mm (Z offset + Z resolution * z)
                        uniformSurface.HeightMap[j][k] = uniformSurfaceMsg.Offset().z + uniformSurfaceMsg.Resolution().z * data;
                    }
                    else
                    {
                        uniformSurface.HeightMap[j][k] = k16S_NULL;
                    }
                }
            }
            std::cout << "Uniform surface data length: " << length << std::endl;
            std::cout << "Uniform surface data width: " << width << std::endl;

            std::cout << (intensityArray ? "Intensity is enabled." : "Intensity is disabled.") << std::endl;
            if (intensityArray)
            {
                std::cout << "Intensity length: " << intensityLength << std::endl;
                std::cout << "Intensity width: " << intensityWidth << std::endl;
            }
            break;
        }

        case GoPxLSdk::MessageType::SURFACE_POINT_CLOUD:
        {

            const GoGdpSurfacePointCloud &pointCloudSurfaceMsg = (const GoGdpSurfacePointCloud &)receivedDataSet.GdpMsgAt(i);
            std::cout << "\nPoint Cloud Surface" << std::endl;
            std::cout << "Date Source: " << pointCloudSurfaceMsg.DataSourceId() << std::endl;
            std::cout << "Array count: " << pointCloudSurfaceMsg.ArrayedCount() << std::endl;
            std::cout << "Array index: " << pointCloudSurfaceMsg.ArrayedIndex() << std::endl;

            // Transform to real world coordinates.
            auto length = pointCloudSurfaceMsg.Length();
            auto width = pointCloudSurfaceMsg.Width();
            auto intensityArray = pointCloudSurfaceMsg.Intensities();
            auto intensityLength = pointCloudSurfaceMsg.IntensityLength();
            auto intensityWidth = pointCloudSurfaceMsg.IntensityWidth();

            surfaceBuffer.resize(length);
            for (k32u row = 0; row < length; row++)
            {
                surfaceBuffer[row] = std::vector<SurfacePoint>(width);
            }

            for (k32u j = 0; j < length; j++)
            {
                for (k32u k = 0; k < width; k++)
                {
                    kPoint3d16s point;
                    kArray2_Item(pointCloudSurfaceMsg.Ranges(), j, k, &point);
                    surfaceBuffer[j][k].x = pointCloudSurfaceMsg.Offset().x + pointCloudSurfaceMsg.Resolution().x * point.x;
                    surfaceBuffer[j][k].y = pointCloudSurfaceMsg.Offset().y + pointCloudSurfaceMsg.Resolution().y * point.y;
                    if (point.z != k16S_NULL)
                    {
                        surfaceBuffer[j][k].z = pointCloudSurfaceMsg.Offset().z + pointCloudSurfaceMsg.Resolution().z * point.z;
                    }
                    else
                    {
                        surfaceBuffer[j][k].z = k16S_NULL;
                    }
                }
            }

            std::cout << "Surface data length: " << length << std::endl;
            std::cout << "Surface data width: " << width << std::endl;

            std::cout << (intensityArray ? "Intensity is enabled." : "Intensity is disabled.") << std::endl;
            if (intensityArray)
            {
                std::cout << "Intensity length: " << intensityLength << std::endl;
                std::cout << "Intensity width: " << intensityWidth << std::endl;
            }
            break;
        }
        }
    }

    // Close the connections.
    getchar();
    gdpClient->Close();
    system.Stop();
    std::cout << "\nSystem stopped." << std::endl;
    system.Disconnect();
    return OK_STATUS;
}

int main(int argc, char **argv)
{
    int status;

    try
    {
        status = ReceiveSurface();
    }
    catch (const GoRequestError &e)
    {
        string errorMessage = "";

        const GoJson &payload = e.GetResponse().Payload();
        if (!payload.Empty() && payload.HasKey("errors"))
        {
            GoJson errors = payload.At("errors");
            errorMessage.append(errors.At("/0/status").ToString());
            errorMessage.append(" ");
            errorMessage.append(errors.At("/0/description").ToString());
        }

        std::cerr << "GoRequestError: " << errorMessage << std::endl;
        std::cerr << "Error sending a REST command to " << e.GetResponse().Path() << std::endl;
        status = ERROR_STATUS;
    }
    catch (const GoChannelError &e)
    {
        std::cerr << "Error: " << e.what() << " - Check sensor status and make sure it is connected." << std::endl;
        status = ERROR_STATUS;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        status = ERROR_STATUS;
    }

    return status;
}
