/**
 * @file    lmi-ext.hpp
 * @brief   Declares C++ workaround.
 *
 * @internal
 * Copyright (C) 2019 by LMI Technologies Inc.
 * Redistributed files must retain the above copyright notice.
 */
#ifndef LMI_EXT_HPP
#define LMI_EXT_HPP

/*
 * Some embedded toolchains do not provide C99 Math support, limiting the capabilities of libstdc++.
 * This file provides a workaround for this limitation by auto-detecting the libstdc++ capabilities
 * and providing alternative implementations for the missing functionality.
 */

#if defined(__GNUC__)
#   include <bits/c++config.h>
#   if !defined(_GLIBCXX_USE_C99)
#       define LMI_EXT_C99_WORKAROUND      (1)
#   endif
#endif

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <sstream>

#include <string>
#include <sstream>
#include <iomanip>
#include <stdlib.h>

// Moved out of the ifdef due to GOS-838
// This template function was moved out of the preprocessor directives to allow for use in unit testing
template <typename T>
T strtoT(const char* str, char** endptr, int base)
{
    std::istringstream stream(str);
    T val;

    stream >> std::setbase(base) >> val;

    if (endptr != nullptr)
    {
        // Sepcial Case where EOF is reached
        if (stream.eof())
        { 
            stream.seekg(0, stream.end);
        }

        auto pos = stream.tellg();
        *endptr = (char*)(str + (int)pos);
    }

    return val;
}

// Workaround for lack of C99 support.

#if defined(LMI_EXT_C99_WORKAROUND)

namespace std
{
    using ::snprintf;

#if !_GLIBCXX_USE_C99_MATH
#undef isfinite
    constexpr bool isfinite(double x)
    {
        return __builtin_isfinite(x);
    }

#undef isinf
    constexpr bool isinf(double x)
    {
        return __builtin_isinf(x);
    }
    
#undef isnan
    constexpr bool isnan(double x)
    {
        return __builtin_isnan(x);
    }

#undef signbit
    constexpr bool signbit(double x)
    {
        return __builtin_signbit(x);
    }
#endif

    template < typename T > std::string to_string(const T& n)
    {
        std::ostringstream stm;
        stm << n;
        return stm.str();
    }

    inline long long int strtoll(const char* str, char** endptr, int base)
    {
        return strtoT<long long int>(str, endptr, base);
    }

    inline unsigned long long int strtoull(const char* str, char** endptr, int base)
    {
        return strtoT<unsigned long long int>(str, endptr, base);
    }

    inline float strtof(const char* str, char** endptr)
    {
        return strtoT<float>(str, endptr, 10);
    }

    inline double strtold(const char* str, char** endptr)
    {
        return strtoT<double>(str, endptr, 10);
    }

    inline int stoi(const string&  str, size_t* idx = 0, int base = 10)
    {
        char* endPos;

        auto ret = (int)strtoll(str.c_str(), &endPos, base);

        if (idx != nullptr)
        {
            *idx = (size_t)(endPos - str.c_str());
        }

        return ret;
    }

    inline long long stoll(const string&  str, size_t* idx = 0, int base = 10)
    {
        char* endPos;

        auto ret = strtoll(str.c_str(), &endPos, base);

        if (idx != nullptr)
        {
            *idx = (size_t)(endPos - str.c_str());
        }

        return ret;
    }

    inline unsigned long long stoull(const string&  str, size_t* idx = 0, int base = 10)
    {
        char* endPos;

        auto ret = strtoull(str.c_str(), &endPos, base);

        if (idx != nullptr)
        {
            *idx = (size_t)(endPos - str.c_str());
        }

        return ret;
    }

    inline long double stold(const string&  str, size_t* idx = 0)
    {
        char* endPos;

        auto ret = (long double)strtold(str.c_str(), &endPos);

        if (idx != nullptr)
        {
            *idx = (size_t)(endPos - str.c_str());
        }

        return ret;
    }
}

#endif

#endif
