/**
 * @file    GoGdpClient.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using GoPxLSdk.GoGdpMsg;
using System;
using System.Collections.Concurrent;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace GoPxLSdk
{
    public class GoGdpClient : IDisposable
    {
        /// <summary>Checks if the client is connected or not.</summary>
        /// <returns>True if connected, false otherwise.</returns>
        public bool IsConnected { get; private set; }

        private const int CONNECT_TIMEOUT_MILLISECONDS = 5000;
        private const int CANCEL_QUERY_INTERVAL_MICROSECONDS = 100000;

        private Action<GoDataSet> _func;

        private bool _async = false;
        private bool _isCompleteSet;

        private TcpClient _tcpClient;
        private GoBinaryReader _reader;
        private GoTimer _timer;
        private Thread _receiveThread;
        private readonly GoDataSet _dataSet;

        private BlockingCollection<GoDataSet> _dataQueue;
        private Thread _dataThread;

        /// <summary>Constructs GoGdpClient.</summary>
        public GoGdpClient()
        {
            _isCompleteSet = false;
            _dataSet = new GoDataSet();
            InitThreadAndTimer();
        }

        private void InitTcpClient()
        {
            try
            {
                _tcpClient = new TcpClient(AddressFamily.InterNetwork)
                {
                    ReceiveTimeout = CONNECT_TIMEOUT_MILLISECONDS
                };
            }
            catch
            {
                _tcpClient?.Dispose();
                throw;
            }
        }

        private void InitThreadAndTimer()
        {
            try
            {
                _receiveThread = new Thread(new ThreadStart(OnDataReceive))
                {
                    IsBackground = true
                };

                _dataThread = new Thread(new ThreadStart(DataThreadEntry))
                {
                    IsBackground = true
                };

                _timer = new GoTimer();
                _dataQueue = new BlockingCollection<GoDataSet>();
            }
            catch
            {
                _dataQueue?.Dispose();

                throw;
            }
        }

        /// <summary>Close the tcp connection.</summary>
        public void Close()
        {
            if (IsConnected)
            {
                IsConnected = false;

                if (_async)
                {
                    _receiveThread.Join();
                    _dataThread.Join();
                }

                _reader?.Dispose();
                _tcpClient?.Dispose();
                _dataQueue?.Dispose();
            }
        }

        /// <summary>Sync Receive data from gdp server.</summary>
        /// <param name="receiveTimeoutMilliseconds">Set time out in milliseconds.</param>
        public void ReceiveDataSync(long receiveTimeoutMilliseconds)
        {
            bool needEndRead = false;
            bool isDataAvailable = false;
            // GOS-5176: temporarily disable data set identifier sanity check until the issue of the surface message
            // kStamp.frame value being different from the stamp/profile messages is resolved.
            // Uncomment this code when the issue is finally resolved.
            //        ulong firstDataSetId = 0;

            if (_isCompleteSet)
            {
                _dataSet.Clear();
                _isCompleteSet = false;
            }

            if (!IsConnected)
            {
                throw new Exception("The GDP server is not connected.");
            }

            _timer.Restart(receiveTimeoutMilliseconds);

            while (IsConnected)
            {
                try
                {
                    if (_timer.IsExpired)
                    {
                        // If the timer has expired due to the client not having data to read, return 
                        // a timeout error. Otherwise, if there is data to read but not enough time to
                        // form the complete data set, return an incomplete error.
                        if (isDataAvailable)
                        {
                            throw new Exception($"{GoStatus.ErrorIncomplete}: Timed out with data still to be read.");
                        }
                        else
                        {
                            throw new Exception($"{GoStatus.ErrorTimeout}: Timed out with no data to read.");
                        }
                    }

                    isDataAvailable = _tcpClient.Client.Poll(CANCEL_QUERY_INTERVAL_MICROSECONDS, SelectMode.SelectRead);

                    if (isDataAvailable)
                    {
                        MessageType type = GetGdpMsgType();
                        needEndRead = true;

                        var msg = GoGdpMsgMake(type);

                        if (msg != null)
                        {
                            msg.Deserialize(_reader);

                            // Get the message set id of the first message in the data set.

                            // GOS-5176: temporarily disable data set identifier sanity check until the issue of the surface message
                            // kStamp.frame value being different from the stamp/profile messages is resolved.
                            // Uncomment this code when the issue is finally resolved.
                            //if (dataSet->Count() == 0)
                            //{
                            //    firstDataSetId = pGoGdpMsg->DataSetId();
                            //}
                            //else 
                            //{
                            //    firstDataSetId = dataSet->GdpMsgAt(0)->DataSetId();
                            //}

                            // If the data set identifier has changed before the last message of the current set could
                            // be reached, discard the incomplete set. Start a new data set with the new 
                            // data set identifier value.
                            // GOS-3905: A message drop count will be added and keep track of the discarded 
                            // messages.
                            //
                            // GOS-5176: temporarily disable data set identifier sanity check until the issue of the surface message
                            // kStamp.frame value being different from the stamp/profile messages is resolved.
                            // Uncomment this code when the issue is finally resolved.
                            //if (pGoGdpMsg->DataSetId() != firstDataSetId)
                            //{
                            //    dataSet->Clear();
                            //    firstDataSetId = pGoGdpMsg->DataSetId();
                            //}

                            _dataSet.Add(msg);

                            // If this is the last message in the data set, then exit function. dataSet
                            // will contain the complete data set.
                            if (msg.IsLastMsg)
                            {
                                // This closes the message type read section.
                                _reader.EndRead();
                                needEndRead = false;
                                _isCompleteSet = true;
                                break;  // Leave while() loop.
                            }
                        }

                        // This closes the message type read section.
                        _reader.EndRead();
                        needEndRead = false;
                    }
                    else
                    {
                        continue;
                    }
                }
                catch(Exception ex)
                {
                    if (needEndRead)
                    {
                        // This closes the message type read section.
                        _reader.EndRead();
                    }

                    throw ex;
                }
            }
        }

        /// <summary>Get the gdp msg type.</summary>
        private MessageType GetGdpMsgType()
        {
            // Header
            _reader.BeginRead<uint>(true);

            try
            {
                ushort messageType = _reader.ReadUInt16();
                return (MessageType)messageType;
            }
            catch
            {
                _reader.EndRead();
                throw;
            }

            // The message type section read remains open (ie. corresponding EndRead() is not called.
        }

        /// <summary>Get the gdp msg dataSet.</summary>
        /// <returns>The dataSet.</returns>
        public GoDataSet DataSet => _dataSet;

        /// <summary>Connect to GdpServer.</summary>
        /// <param name="ipAddr">Ip addresss.</param>
        /// <param name="port">Default DEFAULT_GDP_SERVER_PORT.</param>
        public void Connect(IPAddress ipAddr, int port = Def.GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT)
        {
            if (IsConnected)
            {
                Close();
                IsConnected = false;
            }

            if (!IsConnected)
            {
                InitTcpClient();


                IsConnected = _tcpClient.ConnectAsync(ipAddr, port).Wait(CONNECT_TIMEOUT_MILLISECONDS);
                    //.BeginConnect(ipAddr, port, null, null)
                    //.AsyncWaitHandle.WaitOne(TimeSpan.FromMilliseconds(CONNECT_TIMEOUT_MILLISECONDS));

                if (!IsConnected)
                {
                    throw new TimeoutException($"Tcp client timed out while waiting for connection");
                }
                else
                {
                    try
                    {
                        _reader = new GoBinaryReader(_tcpClient.GetStream());
                    }
                    catch
                    {
                        _reader?.Dispose();
                        throw;
                    }
                }
            }
        }

        private void DataThreadEntry()
        {
            if (_func != null)
            {
                while (IsConnected)
                {
                    try
                    {
                        if (_dataQueue.TryTake(out GoDataSet data, CANCEL_QUERY_INTERVAL_MICROSECONDS / 100))
                        {
                            _func(data);
                        }
                    }
                    catch
                    {
                        // Eat the exception but don't stop the loop/thread
                    }
                }
            }
        }

        /// <summary>Async Receive data from gdp server.</summary>
        /// <param name="func">Callback function to get data.</param>
        public void ReceiveDataAsync(Action<GoDataSet> func)
        {
            this._func = func;

            _receiveThread.Start();
            _dataThread.Start();

            _async = true;
        }

        private void OnDataReceive()
        {
            bool needEndRead = false;
            // GOS-5176: temporarily disable data set identifier sanity check until the issue of the surface message
            // kStamp.frame value being different from the stamp/profile messages is resolved.
            // Uncomment this code when the issue is finally resolved.
            //        ulong firstDataSetId = 0;

            if (!IsConnected)
            {
                return;
            }

            GoDataSet data = new GoDataSet();
            while (IsConnected)
            {
                try
                {
                    if (_tcpClient.Client.Poll(CANCEL_QUERY_INTERVAL_MICROSECONDS, SelectMode.SelectRead))
                    {
                        MessageType type = GetGdpMsgType();
                        needEndRead = true;

                        var msg = GoGdpMsgMake(type);

                        if (msg != null)
                        {
                            msg.Deserialize(_reader);

                            {
                                // GOS-5176: temporarily disable data set identifier sanity check until the issue of the surface message
                                // kStamp.frame value being different from the stamp/profile messages is resolved.
                                // Uncomment this code when the issue is finally resolved.
                                // firstDataSetId = pGoGdpMsg->DataSetId();
                            }

                            // If the data set identifier has changed before the last message of the current set could
                            // be reached, discard the incomplete set. Start a new data set with the new 
                            // data set identifier value.
                            // GOS-3905: A message drop count will be added and keep track of the discarded 
                            // messages.
                            //
                            // GOS-5176: temporarily disable data set identifier sanity check until the issue of the surface message
                            // kStamp.frame value being different from the stamp/profile messages is resolved.
                            // Uncomment this code when the issue is finally resolved.
                            //if (pGoGdpMsg->DataSetId() != firstDataSetId)
                            //{
                            //    data->Clear();
                            //    firstDataSetId = pGoGdpMsg->DataSetId();
                            //}

                            data.Add(msg);

                            // If this is the last message in the set, add the data set to dataQueue. After 
                            // std##move is called, data will be a nullptr and the next message to be received
                            // will be part of a new data set.
                            if (msg.IsLastMsg)
                            {
                                _dataQueue.Add(data);
                                data = new GoDataSet();
                            }
                        }

                        _reader.EndRead();
                        needEndRead = false;
                    }
                    else
                    {
                        continue;
                    }
                }
                catch
                {
                    if (needEndRead)
                    {
                        _reader.EndRead();
                    }

                    //_receiveError = result;
                    return;
                }
            }

            //_receiveError = result;
        }

        private GoGdpMsg.GoGdpMsg GoGdpMsgMake(MessageType type)
        {
            switch (type)
            {
                case MessageType.Stamp:
                    return new GoGdpStamp();
                case MessageType.Image:
                    return new GoGdpImage();
                case MessageType.Signal:
                    return new GoGdpSignal();
                case MessageType.ProfilePointCloud:
                    return new GoGdpProfilePointCloud();
                case MessageType.UniformProfile:
                    return new GoGdpProfileUniform();
                case MessageType.SurfacePointCloud:
                    return new GoGdpSurfacePointCloud();
                case MessageType.UniformSurface:
                    return new GoGdpSurfaceUniform();
                case MessageType.Measurement:
                    return new GoGdpMeasurement();
                case MessageType.NullType:
                    return new GoGdpNull();
                case MessageType.Spots:
                    return new GoGdpSpots();
                case MessageType.Mesh:
                    return new GoGdpMesh();
                case MessageType.Rendering:
                    return new GoGdpRendering();
                case MessageType.PointFeature:
                    return new GoGdpFeaturePoint();
                case MessageType.LineFeature:
                    return new GoGdpFeatureLine();
                case MessageType.PlaneFeature:
                    return new GoGdpFeaturePlane();
                case MessageType.CircleFeature:
                    return new GoGdpFeatureCircle();
                default:
                    return null;
            }
        }

        public void Dispose()
        {
            Close();
        }
    }
}