/**
 * @file    GoGdpGraphics.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System;
using System.Collections.Generic;

namespace GoPxLSdk.GoGdpMsg
{
    public class GoGdpGraphics
    {
        public List<GoPointSet> PointSet { get; } = new List<GoPointSet>();
        public List<GoLineSet> LineSet { get; } = new List<GoLineSet>();
        public List<GoRegion> RegionSet { get; } = new List<GoRegion>();
        public List<GoPlane> PlaneSet { get; } = new List<GoPlane>();
        public List<GoRay> RaySet { get; } = new List<GoRay>();
        public List<GoLabel> LabelSet { get; } = new List<GoLabel>();
        public List<GoPositionObject> PositionSet { get; } = new List<GoPositionObject>();

        /// <summary>Deserialize the graphics object.</summary>
        /// <param name="serializer">The reader to read.</param>
        internal void Deserialize(GoBinaryReader serializer)
        {
            // Read sectioned attribute section containing all graphic counts.
            serializer.BeginRead<ushort>(true);

            ushort pointSetCount;
            ushort lineSetCount;
            ushort regionCount;
            ushort planeCount;
            ushort rayCount;
            ushort labelCount;
            ushort positionCount;

            try
            {
                pointSetCount = serializer.ReadUInt16();
                lineSetCount = serializer.ReadUInt16();
                regionCount = serializer.ReadUInt16();
                planeCount = serializer.ReadUInt16();
                rayCount = serializer.ReadUInt16();
                labelCount = serializer.ReadUInt16();
                positionCount = serializer.ReadUInt16();
            }
            finally
            {
                serializer.EndRead();
            }

            // After the sectioned attribute section contains all the graphic data.
            DeserializePointSets(serializer, pointSetCount);
            DeserializeLineSets(serializer, lineSetCount);
            DeserializeRegions(serializer, regionCount);
            DeserializePlanes(serializer, planeCount);
            DeserializeRays(serializer, rayCount);
            DeserializeLabels(serializer, labelCount);
            DeserializePositions(serializer, positionCount);
        }

        /// <summary>Get the number of point set items in the graphics object.</summary>
        /// <returns>the number of point set items.</returns>
        public long PointSetCount => PointSet.Count;

        /// <summary>Get the specified point set item in the point set list.</summary>
        /// <param name="index">Specify which point set item to return.</param>
        /// <returns>reference to the point set item.</returns>
        public GoPointSet PointSetAt(int index) => PointSet[index];

        /// <summary>Get the number of line set items in the graphics object.</summary>
        /// <returns>the number of line set items.</returns>
        public long LineSetCount => LineSet.Count;

        /// <summary>Get the specified line set item in the line set list.</summary>
        /// <param name="index">Specify which line set item to return.</param>
        /// <returns>reference to the line set item.</returns>
        public GoLineSet LineSetAt(int index) => LineSet[index];

        /// <summary>Get the number of graphics regions in the graphics object.</summary>
        /// <returns>the number of graphics regions.</returns>
        public long RegionCount => RegionSet.Count;

        /// <summary>Get the specified graphics region in the region list.</summary>
        /// <param name="index">Specify which graphics region to return.</param>
        /// <returns>reference to the graphics region.</returns>
        public GoRegion RegionAt(int index) => RegionSet[index];

        /// <summary>Get the number of planes in the graphics object.</summary>
        /// <returns>the number of planes.</returns>
        public long PlaneCount => PlaneSet.Count;

        /// <summary>Get the specified plane in the planes list.</summary>
        /// <param name="index">Specify which plane to return.</param>
        /// <returns>reference to the plane.</returns>
        public GoPlane PlaneAt(int index) => PlaneSet[index];

        /// <summary>Get the number of rays in the graphics object.</summary>
        /// <returns>the number of rays.</returns>
        public long RayCount => RaySet.Count;

        /// <summary>Get the specified ray in the rays list.</summary>
        /// <param name="index">Specify which ray to return.</param>
        /// <returns>reference to the ray.</returns>
        public GoRay RayAt(int index) => RaySet[index];

        /// <summary>Get the number of labels in the graphics object.</summary>
        /// <returns>the number of labels.</returns>
        public long LabelCount => LabelSet.Count;

        /// <summary>Get the specified label in the labels list.</summary>
        /// <param name="index">Specify which label to return.</param>
        /// <returns>reference to the label.</returns>
        public GoLabel LabelAt(int index) => LabelSet[index];

        /// <summary>Get the number of positions in the graphics object.</summary>
        /// <returns>the number of positions.</returns>
        public long PositionCount => PositionSet.Count;

        /// <summary>Get the specified position in the positions list.</summary>
        /// <param name="index">Specify which position to return.</param>
        /// <returns>reference to the position.</returns>
        public GoPositionObject PositionAt(int index) => PositionSet[index];

        // Helper functions to deserialize the different bytes of graphics types.
        private void DeserializePointSets(GoBinaryReader serializer, ushort count)
        {
            PointSet.Clear();

            for (int i = 0; i < count; i++)
            {
                GoPointSet pointSetEntry = new GoPointSet();
                pointSetEntry.Deserialize(serializer);
                PointSet.Add(pointSetEntry);
            }
        }

        private void DeserializeLineSets(GoBinaryReader reader, ushort count)
        {
            LineSet.Clear();

            for (int i = 0; i < count; i++)
            {
                GoLineSet lineSetEntry = new GoLineSet(reader);
                LineSet.Add(lineSetEntry);
            }
        }

        private void DeserializeRegions(GoBinaryReader reader, ushort count)
        {
            RegionSet.Clear();

            for (int i = 0; i < count; i++)
            {
                RegionType type = (RegionType)reader.ReadByte();

                switch (type)
                {
                    case RegionType.ProfileRegion2d:
                        {
                            var region = new ProfileRegion2d(reader);
                            RegionSet.Add(region);
                            break;
                        }
                    case RegionType.SurfaceRegion2d:
                        {
                            var region = new SurfaceRegion2d(reader);
                            RegionSet.Add(region);
                            break;
                        }
                    case RegionType.Region3d:
                        {
                            var region = new Region3d(reader);
                            RegionSet.Add(region);
                            break;
                        }
                    default:
                        throw new NotSupportedException($"{nameof(GoGdpRendering)} does not supportdeserializing region with type {type}");
                }
            }
        }

        private void DeserializePlanes(GoBinaryReader reader, ushort count)
        {
            PlaneSet.Clear();

            for (int i = 0; i < count; i++)
            {
                GoPlane planeEntry = new GoPlane(reader);
                PlaneSet.Add(planeEntry);
            }
        }
        private void DeserializeRays(GoBinaryReader reader, ushort count)
        {
            RaySet.Clear();

            for (int i = 0; i < count; i++)
            {
                GoRay rayEntry = new GoRay(reader);
                RaySet.Add(rayEntry);
            }
        }

        private void DeserializeLabels(GoBinaryReader reader, ushort count)
        {
            LabelSet.Clear();

            for (int i = 0; i < count; i++)
            {
                GoLabel labelEntry = new GoLabel(reader);
                LabelSet.Add(labelEntry);
            }
        }

        private void DeserializePositions(GoBinaryReader reader, ushort count)
        {
            PositionSet.Clear();

            for (int i = 0; i < count; i++)
            {
                GoPositionObject positionEntry = new GoPositionObject(reader);
                PositionSet.Add(positionEntry);
            }
        }
    }
}
