/**
 * @file    GoGdpMesh.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System.Collections.Generic;

namespace GoPxLSdk.GoGdpMsg
{
    public enum ChannelID
    {
        Vertex = 0,
        Facet = 1,
        FacetNormal = 2,
        VertexNormal = 3,
        VertexTexture = 4,
        VertexCurvature = 5,
    };

    public enum ChannelType
    {
        Invalid = 0,
        Vertex = 1,
        Facet = 2,
        FacetNormal = 3,
        VertexNormal = 4,
        VertexTexture = 5,
        VertexCurvature = 6
    };

    public enum ChannelState
    {
        Error = 0,
        Unallocated = 1,
        Allocated = 2,
        Empty = 3,
        PartialUsed = 4,
        Full = 5
    };

    public class GoFacet32u
    {
        public uint Vertex1 { get; }
        public uint Vertex2 { get; }
        public uint Vertex3 { get; }

        public GoFacet32u(uint vertex1, uint vertex2, uint vertex3)
        {
            Vertex1 = vertex1;
            Vertex2 = vertex2;
            Vertex3 = vertex3;
        }
    }

    public class MeshMsgChannel
    {
        public ChannelID Id { get; }
        public ChannelType Type { get; }
        public ChannelState State { get; }
        public uint Flag { get; }
        public uint AllocatedCount { get; }
        public uint UsedCount { get; }
        public List<object> Buffer { get; } = new List<object>();

        public MeshMsgChannel(ChannelID id, ChannelType type, ChannelState state, uint flag, uint allocatedCount, uint usedCount)
        {
            Id = id;
            Type = type;
            State = state;
            Flag = flag;
            AllocatedCount = allocatedCount;
            UsedCount = usedCount;
        }
    }

    public class GoGdpMesh : GoGdpMsg
    {
        /// <summary>Indicator whether any data are allocated</summary>
        /// <returns>true if allocated.</returns>
        public bool HasData { get; internal set; } = false;

        /// <summary>Number of system channels(currently 6)</summary>
        /// <returns>true if allocated.</returns>
        public uint SystemChannelCount { get; internal set; }

        /// <summary>Maximum number of user channels supported (currently 5)</summary>
        /// <returns>maximum number of user channels supported.</returns>
        public uint MaxUserChannelCount { get; internal set; }

        /// <summary>Number of currently active user channels.</summary>
        /// <returns>number of currently active user channels.</returns>
        public uint UserChannelCount { get; internal set; }

        /// <summary>Number of all channels, including all system and active user channels.</summary>
        /// <returns>number of all channels.</returns>
        public uint ChannelCount { get; internal set; }

        /// <summary>Offset of mesh data.</summary>
        /// <returns>offset values of mesh data.</returns>
        public GoPoint3d64f Offset { get; internal set; } = new GoPoint3d64f(0, 0, 0);

        /// <summary>Range of mesh data.</summary>
        /// <returns>range values of mesh data.</returns>
        public GoPoint3d64f Range { get; internal set; } = new GoPoint3d64f(0, 0, 0);

        /// <summary>Get channel type.</summary>
        /// <param name="id">The channel id which to get type value.</param>
        /// <returns>channel type value.</returns>
        public ChannelType ChannelType(int id) => Channels[id].Type;

        /// <summary>Get channel state.</summary>
        /// <param name="id">The channel id which get state value.</param>
        /// <returns>channel state value.</returns>
        public ChannelState ChannelState(int id) => Channels[id].State;

        /// <summary>
        /// Channel flag is an user specified field. 
        /// It can be used to send any optional data
        /// to determine how to deserialize user channel data.
        /// </summary>
        /// <param name="id">The channel id which to get flag.</param>
        /// <returns>channel flag value.</returns>
        public uint ChannelFlag(int id) => Channels[id].Flag;

        /// <summary>
        /// Number of allocated channel data items. 
        /// System channels – Equal to number of items(ie, allocateCount for vertex channel is number of Point3d32f allocated).
        /// User channels – Equal to number of bytes(8u) of entire buffer.
        /// </summary>
        /// <param name="id">The channel id which to get allocated channel data count.</param>
        /// <returns>size of channel count.</returns>
        public uint AllocatedChannelDataCount(int id) => Channels[id].AllocatedCount;

        /// <summary>Number of used channel data items.</summary>
        /// <param name="id">The channel id which to get used channel data count.</param>
        /// <returns>number of used channel data items.</returns>
        public uint UsedChannelDataCount(int id) => Channels[id].UsedCount;

        /// <summary>Get data buffer of various types depending on channel ID.</summary>
        /// <param name="id">The channel id which to get the buffer data.</param>
        /// <returns>array of data values.</returns>
        public List<object> ChannelData(int id) => Channels[id].Buffer;

        public List<MeshMsgChannel> Channels { get; } = new List<MeshMsgChannel>();

        /// <summary>Constructs class GoGdpMesh.</summary>
        internal GoGdpMesh()
            : base(MessageType.Mesh)
        { }

        /// <summary>Deserialize mesh message.</summary>
        /// <param name="serializer">The serializer to read.</param>
        internal override void Deserialize(GoBinaryReader serializer)
        {
            bool needEndRead = false;

            base.Deserialize(serializer);

            serializer.BeginRead<ushort>(true);
            needEndRead = true;

            try
            {
                HasData = serializer.ReadBool();
                SystemChannelCount = serializer.ReadUInt32();
                MaxUserChannelCount = serializer.ReadUInt32();
                UserChannelCount = serializer.ReadUInt32();
                ChannelCount = serializer.ReadUInt32();

                double offsetX = serializer.ReadDouble();
                double offsetY = serializer.ReadDouble();
                double offsetZ = serializer.ReadDouble();
                Offset = new GoPoint3d64f(offsetX, offsetY, offsetZ);

                double rangeX = serializer.ReadDouble();
                double rangeY = serializer.ReadDouble();
                double rangeZ = serializer.ReadDouble();
                Range = new GoPoint3d64f(rangeX, rangeY, rangeZ);

                serializer.EndRead();
                needEndRead = false;

                Channels.Clear();

                for (uint i = 0; i < ChannelCount; i++)
                {
                    serializer.BeginRead<ushort>(true);
                    needEndRead = true;
                    uint id = serializer.ReadUInt32();
                    uint type = serializer.ReadUInt32();
                    int state = serializer.ReadInt32();
                    uint flag = serializer.ReadUInt32();
                    uint allocatedSize = serializer.ReadUInt32();
                    uint usedSize = serializer.ReadUInt32();
                    serializer.EndRead();
                    needEndRead = false;

                    var channel = new MeshMsgChannel((ChannelID)id, (ChannelType)type, (ChannelState)state, flag, allocatedSize, usedSize);
                    if (channel.AllocatedCount > 0)
                    {
                        switch (channel.Id)
                        {
                            case ChannelID.Vertex:
                            case ChannelID.FacetNormal:
                            case ChannelID.VertexNormal:
                                {
                                    for (int j = 0; j < channel.AllocatedCount; j++)
                                    {
                                        float x = serializer.ReadSingle();
                                        float y = serializer.ReadSingle();
                                        float z = serializer.ReadSingle();

                                        channel.Buffer.Add(new GoPoint3d32f(x, y, z));
                                    }
                                    break;
                                }
                            case ChannelID.Facet:
                                {
                                    for (int j = 0; j < channel.AllocatedCount; j++)
                                    {
                                        uint vertex1 = serializer.ReadUInt32();
                                        uint vertex2 = serializer.ReadUInt32();
                                        uint vertex3 = serializer.ReadUInt32();
                                        channel.Buffer.Add(new GoFacet32u(vertex1, vertex2, vertex3));
                                    }
                                    break;
                                }
                            case ChannelID.VertexCurvature:
                                {
                                    for (int j = 0; j < channel.AllocatedCount; j++)
                                    {
                                        float item = serializer.ReadSingle();
                                        channel.Buffer.Add(item);
                                    }
                                    break;
                                }
                            case ChannelID.VertexTexture:
                            default:
                                for (int j = 0; j < channel.AllocatedCount; j++)
                                {
                                    byte item = serializer.ReadByte();
                                    channel.Buffer.Add(item);
                                }
                                break;
                        }
                    }

                    Channels.Add(channel);
                }
            }
            finally
            {
                if(needEndRead)
                {
                    serializer.EndRead();
                }
            }
        }
    }
}