/**
 * @file    GoGdpRendering.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System;
using System.Collections.Generic;

namespace GoPxLSdk.GoGdpMsg
{
    public class GoPointSet
    {
        public float Size { get; private set; }
        public uint Color { get; private set; }
        public int Shape { get; private set; }
        public IReadOnlyList<GoPoint3d32f> Points => _points;

        private readonly List<GoPoint3d32f> _points = new List<GoPoint3d32f>();

        public GoPointSet()
        { }

        public GoPointSet(float size, uint color, int shape, List<GoPoint3d32f> points)
        {
            Size = size;
            Color = color;
            Shape = shape;
            _points = points;
        }

        internal void Deserialize(GoBinaryReader serializer)
        {
            ushort pointCount;
            _points.Clear();

            try
            {
                serializer.BeginRead<ushort>(true);

                Size = serializer.ReadSingle();
                Color = serializer.ReadUInt32();
                Shape = serializer.ReadInt32();
                pointCount = serializer.ReadUInt16();
            }
            finally
            {
                serializer.EndRead();
            }

            for (int i = 0; i < pointCount; i++)
            {
                var p = serializer.ReadGoPoint3d32f();
                _points.Add(new GoPoint3d32f(p.X, p.Y, p.Z));
            }
        }
    }

    public class GoLineSet
    {
        public float Width { get; }
        public uint Color { get; }
        public bool HasStartPointArrow { get; }
        public bool HasEndPointArrow { get; }
        public IReadOnlyList<GoPoint3d32f> Points => _pointsList;

        private readonly List<GoPoint3d32f> _pointsList = new List<GoPoint3d32f>();

        public GoLineSet(float width, uint color, bool hasStartPointArrow, bool hasEndPointArrow, List<GoPoint3d32f> points)
        {
            Width = width;
            Color = color;
            HasStartPointArrow = hasStartPointArrow;
            HasEndPointArrow = hasEndPointArrow;
            _pointsList = points;
        }

        internal GoLineSet(GoBinaryReader reader)
        {
            ushort pointCount;

            try
            {
                reader.BeginRead<ushort>(true);

                Width = reader.ReadSingle();
                Color = reader.ReadUInt32();

                HasStartPointArrow = reader.ReadBool();
                HasEndPointArrow = reader.ReadBool();

                pointCount = reader.ReadUInt16();

            }
            finally
            {
                reader.EndRead();
            }

            for (int j = 0; j < pointCount; j++)
            {
                var x = reader.ReadSingle();
                var y = reader.ReadSingle();
                var z = reader.ReadSingle();

                _pointsList.Add(new GoPoint3d32f(x, y, z));
            }
        }
    }

    public enum RegionType : byte
    {
        ProfileRegion2d = 0,
        Region3d = 1,
        SurfaceRegion2d = 2,  // For future use.
    };

    public abstract class GoRegion
    {
        public RegionType Type { get; }

        protected GoRegion(RegionType type)
        {
            Type = type;
        }
    }

    public class ProfileRegion2d : GoRegion
    {
        public double X { get; }
        public double Z { get; }
        public double Width { get; }
        public double Height { get; }
        public double AngleY { get; }

        public ProfileRegion2d(double x, double z, double width, double height, double angleY)
            : base(RegionType.ProfileRegion2d)
        {
            X = x;
            Z = z;
            Width = width;
            Height = height;
            AngleY = angleY;
        }

        internal ProfileRegion2d(GoBinaryReader reader)
            : base(RegionType.ProfileRegion2d)
        {
            try
            {
                reader.BeginRead<ushort>(true);

                X = reader.ReadDouble();
                Z = reader.ReadDouble();
                Width = reader.ReadDouble();
                Height = reader.ReadDouble();
                AngleY = reader.ReadDouble();
            }
            finally
            {
                reader.EndRead();
            }
        }
    }

    public class SurfaceRegion2d : GoRegion
    {
        public double X { get; }
        public double Y { get; }
        public double Width { get; }
        public double Length { get; }
        public double AngleZ { get; }

        public SurfaceRegion2d(double x, double y, double width, double length, double angleZ)
            : base(RegionType.SurfaceRegion2d)
        {
            X = x;
            Y = y;
            Width = width;
            Length = length;
            AngleZ = angleZ;
        }

        internal SurfaceRegion2d(GoBinaryReader reader)
            : base(RegionType.SurfaceRegion2d)
        {
            try
            {
                reader.BeginRead<ushort>(true);

                X = reader.ReadDouble();
                Y = reader.ReadDouble();
                Width = reader.ReadDouble();
                Length = reader.ReadDouble();
                AngleZ = reader.ReadDouble();
            }
            finally
            {
                reader.EndRead();
            }
        }
    }

    public class Region3d : GoRegion
    {
        public double X { get; }
        public double Y { get; }
        public double Z { get; }
        public double Width { get; }
        public double Length { get; }
        public double Height { get; }
        public double AngleZ { get; }

        public Region3d(double x, double y, double z, double width, double length, double height, double angleZ)
            : base(RegionType.Region3d)
        {
            X = x;
            Y = y;
            Z = z;
            Width = width;
            Length = length;
            Height = height;
            AngleZ = angleZ;
        }

        internal Region3d(GoBinaryReader reader)
            : base(RegionType.Region3d)
        {
            try
            {
                reader.BeginRead<ushort>(true);

                X = reader.ReadDouble();
                Y = reader.ReadDouble();
                Z = reader.ReadDouble();
                Width = reader.ReadDouble();
                Length = reader.ReadDouble();
                Height = reader.ReadDouble();
                AngleZ = reader.ReadDouble();
            }
            finally
            {
                reader.EndRead();
            }
        }
    }

    public class GoPlane
    {
        public float Distance { get; }
        public GoPoint3d32f Normal { get; }


        public GoPlane(float distance, float normalX, float normalY, float normalZ)
            : this(distance, new GoPoint3d32f(normalX, normalY, normalZ))
        { }

        public GoPlane(float distance, GoPoint3d32f normal)
        {
            Distance = distance;
            Normal = normal;
        }

        internal GoPlane(GoBinaryReader reader)
        {
            try
            {
                reader.BeginRead<ushort>(true);

                Distance = reader.ReadSingle();
                var x = reader.ReadSingle();
                var y = reader.ReadSingle();
                var z = reader.ReadSingle();

                Normal = new GoPoint3d32f(x, y, z);
            }
            finally
            {
                reader.EndRead();
            }
        }
    }

    public class GoRay
    {
        public GoPoint3d32f Position { get; }
        public GoPoint3d32f Direction { get; }
        public float Width { get; }
        public uint Color { get; }

        public GoRay(float posX, float posY, float posZ, float dirX, float dirY, float dirZ, float width, uint color)
            : this(new GoPoint3d32f(posX, posY, posZ), new GoPoint3d32f(dirX, dirY, dirZ), width, color)
        { }

        public GoRay(GoPoint3d32f position, GoPoint3d32f direction, float width, uint color)
        {
            Position = position;
            Direction = direction;
            Width = width;
            Color = color;
        }

        internal GoRay(GoBinaryReader reader)
        {
            try
            {
                reader.BeginRead<ushort>(true);

                var posX = reader.ReadSingle();
                var posY = reader.ReadSingle();
                var posZ = reader.ReadSingle();
                Position = new GoPoint3d32f(posX, posY, posZ);

                var dirX = reader.ReadSingle();
                var dirY = reader.ReadSingle();
                var dirZ = reader.ReadSingle();
                Direction = new GoPoint3d32f(dirX, dirY, dirZ);

                Width = reader.ReadSingle();
                Color = reader.ReadUInt32();
            }
            finally
            {
                reader.EndRead();
            }
        }
    }

    public class GoLabel
    {
        public string Text { get; }
        public GoPoint3d64f Position { get; }

        public GoLabel(string text, double x, double y, double z)
            : this(text, new GoPoint3d64f(x, y, z))
        { }

        public GoLabel(string text, GoPoint3d64f position)
        {
            Text = text;
            Position = position;
        }

        internal GoLabel(GoBinaryReader reader)
        {
            try
            {
                reader.BeginRead<ushort>(true);

                var length = reader.ReadUInt16();
                Text = Utils.ReadText(reader, length);

                var x = reader.ReadDouble();
                var y = reader.ReadDouble();
                var z = reader.ReadDouble();
                Position = new GoPoint3d64f(x, y, z);
            }
            finally
            {
                reader.EndRead();

            }
        }
    }

    public class GoPositionObject
    {
        public GoPoint3d64f Position { get; }
        public byte Type { get; }

        public GoPositionObject(double x, double y, double z, byte type)
            : this(new GoPoint3d64f(x, y, z), type)
        { }

        public GoPositionObject(GoPoint3d64f position, byte type)
        {
            Position = position;
            Type = type;
        }

        internal GoPositionObject(GoBinaryReader reader)
        {
            try
            {
                reader.BeginRead<ushort>(true);

                var x = reader.ReadDouble();
                var y = reader.ReadDouble();
                var z = reader.ReadDouble();
                Position = new GoPoint3d64f(x, y, z);

                Type = reader.ReadByte();
            }
            finally
            {
                reader.EndRead();
            }
        }
    }

    public class GoGdpRendering : GoGdpMsg
    {
        /// <summary>Deserialize rendering msg.</summary>
        /// <param name="reader">The serializer to read.</param>
        internal override void Deserialize(GoBinaryReader reader)
        {
            base.Deserialize(reader);
            Graphics.Deserialize(reader);
        }

        /// <summary>Get the graphics object in the rendering message.</summary>
        /// <returns>the graphics object reference.</returns>
        public GoGdpGraphics Graphics { get; private set; } = new GoGdpGraphics();

        /// <summary>Constructs GoGdpRendering.</summary>
        public GoGdpRendering()
            : base(MessageType.Rendering)
        { }
    }
}