/**
 * @file    MessageType.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
namespace GoPxLSdk.GoGdpMsg
{
    /// <summary>
    /// List of enums representing types available to be serialized out through GDP
    /// This list is essentially the same as GoCore::MessageType.
    ///
    /// Although the list is similar to PipeDataType in GoPipe::PipeDataType,
    /// The key difference is that MessageType represents types to be serialized out.
    /// These enums should be kept as close to each other as possible.
    /// </summary>
    public enum MessageType : ushort
    {
        //---------------------------------------------------------------------
        // Control types
        //---------------------------------------------------------------------

        /// <summary>
        /// Signals to clients that data on a stream is invalidated.
        /// </summary>
        Signal = 1,

        // Reserved for control type expansion.
        //---------------------------------------------------------------------
        // Common data types.
        //---------------------------------------------------------------------

        /// <summary>
        /// Null data type that contains a status code. 
        /// </summary>
        NullType = 10,

        /// <summary>
        /// Stamp information.
        /// </summary>
        Stamp = 11,

        /// <summary>
        /// Uniform profile data.
        /// </summary>
        UniformProfile = 12,

        /// <summary>
        /// Raw profile data.
        /// </summary>
        ProfilePointCloud = 13,

        /// <summary>
        /// Uniform surface data.
        /// </summary>
        UniformSurface = 14,

        /// <summary>
        /// Raw surface data.
        /// </summary>
        SurfacePointCloud = 15,

        /// <summary>
        /// Image data.
        /// </summary>
        Image = 16,

        /// <summary>
        /// Spot data associated with the image.
        /// </summary>
        Spots = 17,

        /// <summary>
        /// Mesh data.
        /// </summary>
        Mesh = 18,

        /// <summary>
        /// Measurement data.
        /// </summary>
        Measurement = 19,

        // Reserved for common data type expansion.
        //---------------------------------------------------------------------
        // Types containing graphical and feature data.
        //---------------------------------------------------------------------

        /// <summary>
        /// Graphical data associated with an output.
        /// </summary>
        Rendering = 70,

        /// <summary>
        /// Point feature data.
        /// </summary>
        PointFeature = 71,

        /// <summary>
        /// Line feature data.
        /// </summary>
        LineFeature = 72,

        /// <summary>
        /// Plane feature data.
        /// </summary>
        PlaneFeature = 73,

        /// <summary>
        ///  Circle feature data.
        /// </summary>
        CircleFeature = 74,

        // Reserved for feature data type expansion.
        //---------------------------------------------------------------------
        // Serialization only types. (not a framework type)
        //---------------------------------------------------------------------
        Health = 100
    };
}