/**
 * @file    GoRestClient.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using GoPxLSdk.Internal;
using Newtonsoft.Json.Linq;
using System.Collections.Generic;
using System.Net;

namespace GoPxLSdk
{
    public class GoRestClient
    {
        /// <summary>Checks if the REST Client is connected or not.</summary>
        /// <returns>True if connected, false otherwise.</returns>
        public bool IsConnected => _channel.IsConnected;

        // Locks when GoRestClient instance command is being used by a thread.
        private readonly object _commandMutex = new object();

        private readonly Queue<GoTransaction> _transactionQueue = new Queue<GoTransaction>();
        private readonly IGoCtrlChannel _channel;

        private GoStreamHandler _streamHandler;
        private GoNotificationHandler _notificationHandler;

        public GoRestClient()
        {
            _channel = new Internal.V1.GoCtrlChannelV1();
            _channel.SetResponseHandler(ResponseHandler);
        }

        /// <summary>
        /// Notification handler used in GoRestClient::Sub.
        /// The notification object is passed into the receiving function.
        /// </summary>
        /// <param name="response"></param>
        public delegate void GoNotificationHandler(GoNotificationResponse response);

        /// <summary>
        /// Data handler used in GoRestClient::StartStream.
        /// The stream message object is passed into the receiving function.
        /// </summary>
        /// <param name="response"></param>
        public delegate void GoStreamHandler(GoStreamResponse response);

        /// <summary>Connects to a REST endpoint at the provided IP address and port. Waits until the connection
        /// is established.</summary>
        /// <param name="address">The endpoint's IP address.</param>
        /// <param name="port">The endpoint's port. If 0, the default port will be used.</param>
        public void Connect(IPAddress address, uint port = 0)
        {
            lock (_commandMutex)
            {
                _channel.Connect(address, (int)port);
            }
        }

        /// <summary>Disconnects from a REST endpoint. Waits until the disconnection is complete.</summary>
        public void Disconnect()
        {
            lock (_commandMutex)
            {
                _channel.Disconnect();
            }
        }

        /// <summary>Reads a resource and returns its data. Analogous to HTTP's GET method: there is no
        /// content/body, and 'args' are similar to query strings.</summary>
        /// <param name="uri">The URI to the resource.</param>
        /// <param name="content">JObject object of the content for the request.</param>
        /// <param name="args">JObject object of arguments for the request.</param>
        /// <returns>Transaction object to monitor status and wait for results.</returns>
        public GoTransaction Read(string uri, JObject content = null, JObject args = null)
        {
            return TransactionHelper(GoRequestMethod.Read, uri, content, args);
        }

        /// <summary>Updates a resource with the provided content. Analogous to HTTP's PUT method: there
        /// is content/body, and 'args' are similar to query strings.</summary>
        /// <param name="uri">The URI to the resource.</param>
        /// <param name="content">JObject object of the content for the request.</param>
        /// <param name="args">JObject object of arguments for the request.</param>
        /// <returns>Transaction object to monitor status and wait for results.</returns>
        public GoTransaction Update(string uri, JObject content = null, JObject args = null)
        {
            return TransactionHelper(GoRequestMethod.Update, uri, content, args);
        }

        /// <summary>Creates a resource with the provided content. Analogous to HTTP's POST method: there
        /// is content/body, and 'args' are similar to query strings.</summary>
        /// <param name="uri">The URI to the resource.</param>
        /// <param name="content">JObject object of the content for the request.</param>
        /// <param name="args">JObject object of arguments for the request.</param>
        /// <returns>Transaction object to monitor status and wait for results.</returns>
        public GoTransaction Create(string uri, JObject content = null, JObject args = null)
        {
            return TransactionHelper(GoRequestMethod.Create, uri, content, args);
        }

        /// <summary>Deletes a resource. Analogous to HTTP's DELETE method: there is no content/body, and
        /// 'args' are similar to query strings.</summary>
        /// <param name="uri">The URI to the resource.</param>
        /// <param name="content">JObject object of the content for the request.</param>
        /// <param name="args">JObject object of arguments for the request.</param>
        /// <returns>Transaction object to monitor status and wait for results.</returns>
        public GoTransaction Delete(string uri, JObject content = null, JObject args = null)
        {
            return TransactionHelper(GoRequestMethod.Delete, uri, content, args);
        }

        /// <summary>Calls a callable resource. Not all resources are callable, check the API docs.</summary>
        /// <param name="uri">The URI to the resource.</param>
        /// <param name="content">JObject object of the content for the request.</param>
        /// <param name="args">JObject object of arguments for the request.</param>
        /// <returns>Transaction object to monitor status and wait for results</returns>
        public GoTransaction Call(string uri, JObject content = null, JObject args = null)
        {
            return TransactionHelper(GoRequestMethod.Call, uri, content, args);
        }

        /// <summary>Subscribes for notifications on an observable resource. Not all resources are observable,
        /// check the API docs. Ensure you have set the subscription handler before calling
        /// this function to ensure you do not miss any messages.</summary>
        /// <param name="uri">The URI to the resource.</param>
        /// <param name="content">JObject object of the content for the request.</param>
        /// <param name="args">JObject object of arguments for the request.</param>
        /// <returns>Transaction object to monitor status and wait for results</returns>
        public GoTransaction Sub(string uri, JObject content = null, JObject args = null)
        {
            return TransactionHelper(GoRequestMethod.Sub, uri, content, args);
        }

        /// <summary>Unsubscribes for notifications on an observable resource. Not all resources are observable,
        /// check the API docs.</summary>
        /// <param name="uri">The URI to the resource.</param>
        /// <param name="content">JObject object of the content for the request.</param>
        /// <param name="args">JObject object of arguments for the request.</param>
        /// <returns>Transaction object to monitor status and wait for results</returns>
        public GoTransaction UnSub(string uri, JObject content = null, JObject args = null)
        {
            return TransactionHelper(GoRequestMethod.UnSub, uri, content, args);
        }

        /// <summary>Start streaming a streamable resource. Not all resources are streamable, check the
        /// API docs. Ensure you have set the stream handler before calling this function
        /// to ensure you do not miss any messages.</summary>
        /// <param name="uri">The URI to the resource.</param>
        /// <param name="content">JObject object of the content for the request.</param>
        /// <param name="args">JObject object of arguments for the request.</param>
        /// <returns>Transaction object to monitor status and wait for results</returns>
        public GoTransaction StartStream(string uri, JObject content = null, JObject args = null)
        {
            return TransactionHelper(GoRequestMethod.Stream, uri, content, args);
        }

        /// <summary>Stops streaming a streamable resource. Not all resources are streamable, check the
        /// API docs.</summary>
        /// <param name="uri">The URI to the resource.</param>
        /// <param name="content">JObject object of the content for the request.</param>
        /// <param name="args">JObject object of arguments for the request.</param>
        /// <returns>Transaction object to monitor status and wait for results</returns>
        public GoTransaction StopStream(string uri, JObject content = null, JObject args = null)
        {
            return TransactionHelper(GoRequestMethod.CancelStream, uri, content, args);
        }

        /// <summary>Sets the callback handler to receive subscription notifications.</summary>
        /// <param name="callback">The callback which receives subscription notifications</param>
        public void SetStreamHandler(GoStreamHandler callback)
        {
            lock (_commandMutex)
            {
                _streamHandler = callback;
            }
        }

        /// <summary>Sets the callback handler to receive subscription notifications.</summary>
        /// <param name="callback">The callback which receives subscription notifications</param>
        public void SetSubHandler(GoNotificationHandler callback)
        {
            lock (_commandMutex)
            {
                _notificationHandler = callback;
            }
        }

        private void ResponseHandler(GoResponse response)
        {
            lock (_commandMutex)
            {
                if (response == null)
                {
                    return;
                }

                // TODO: Response messages have Type set to Request
                // in GoResponse.cs.
                // What is the purpose of this filtering of Request msg?
                // What does Callback() do?
                if (response.Type != GoResponseType.Request)
                {
                    Callback(response);
                    return;
                }

                if (_transactionQueue.Count > 0)
                {
                    var transaction = _transactionQueue.Dequeue();
                    transaction.OnResponse((GoRequestResponse)response);
                }
            }
        }

        private void Callback(GoResponse response)
        {
            switch (response.Type)
            {
                case GoResponseType.Notification:
                    _notificationHandler?.Invoke((GoNotificationResponse)response);
                    break;
                case GoResponseType.Stream:
                    _streamHandler?.Invoke((GoStreamResponse)response);
                    break;
                default:
                    break;
            }
        }

        private GoTransaction TransactionHelper(GoRequestMethod method, string uri, JObject content = null, JObject args = null)
        {
            var request = new GoRequest(method, uri, content, args);
            var transaction = new GoTransaction();

            lock (_commandMutex)
            {
                _transactionQueue.Enqueue(transaction);
                _channel.SendRequest(request);

                return transaction;
            }
        }
    }
}