/**
 * @file    GoTransaction.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System.Threading.Tasks;

namespace GoPxLSdk
{
    public class GoTransaction
    {
        private const int TRANSACTION_TIMEOUT_MILLISEC = 3000;

        private readonly TaskCompletionSource<GoRequestResponse> _promise;

        /// <summary>Transaction handle to simplify working with asynchronous result data.</summary>
        internal GoTransaction()
            :base()
        {
            _promise = new TaskCompletionSource<GoRequestResponse>();
        }

        /// <summary>Gets the future which stores the transaction promise.</summary>
        /// <param name="response">Response object</param>
        /// <returns>The Transaction Promise which holds the response.</returns>
        internal void OnResponse(GoRequestResponse response)
        {
            _promise.SetResult(response);
        }

        /// <summary>
        /// Checks the future transaction response value without returning it. Waits until the 
        /// response is available.
        /// </summary>
        /// <param name="timeout">Time to wait for the response in milliseconds. Default value is <see cref="TRANSACTION_TIMEOUT_MILLISEC"/>.</param>
        public void CheckResponse(int timeout = TRANSACTION_TIMEOUT_MILLISEC)
        {
            bool completed = _promise.Task.Wait(timeout);
            if (!completed)
            {
                throw new GoChannelException();
            }

            GoRequestResponse response = _promise.Task.Result;
            if(response.Status != GoStatus.Ok)
            {
                throw new GoRequestException(response);
            }
        }

        /// <summary>Gets the future transaction response value. Waits until the response is available.</summary>
        /// <param name="timeout">Time to wait for the response in milliseconds.</param>
        /// <returns>The request's response.</returns>
        public GoRequestResponse GetResponse(int timeout = TRANSACTION_TIMEOUT_MILLISEC)
        {
            CheckResponse(timeout);
            return _promise.Task.Result;
        }
    }
}