/**
 * @file    IGoCtrlChannel.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using System;
using System.Net;

namespace GoPxLSdk.Internal
{
    /// <summary>
    /// Function to receive responses from the server asynchronously.
    /// </summary>
    public delegate void GoCtrlChannelResponseHandler(GoResponse response);

    /// <summary>
    /// Interface for all control channel implementations.
    /// Currently only implemented by GoCtrlChannelV1.
    /// </summary>
    public interface IGoCtrlChannel : IDisposable
    {
        void Connect(IPAddress address, int port, int connectionTimeout = Def.DEFAULT_TCP_TIMEOUT_MILLISECONDS);

        /// <summary>Disconnects from the server.</summary>
        void Disconnect();

        /// <summary>Checks if connected to the server.</summary>
        /// <returns>true if connected</returns>
        bool IsConnected { get; }

        /// <summary>
        /// Sends a request according to the implemented carrier protocol.
        /// Should store the raw, complete request json on the GoRequest object in case this object
        /// is cached and possible reviewed for debugging purposes.
        /// </summary>
        /// <exception cref="GoChannelException">Thrown if not connected.</exception>
        void SendRequest(GoRequest request);

        /// <summary>
        /// Sets a callback to receive responses from the server, using the implemented carrier protocol.
        /// </summary>
        void SetResponseHandler(GoCtrlChannelResponseHandler callback);
    }
}