/**
 * @file    EnumAsLowerCaseStringFormatter.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 */
using MessagePack.Formatters;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.Serialization;
using MessagePack;

namespace GoPxLSdk
{
    // Variant for forced lower case of Enum names. This can be deleted and replaced with [EnumMember] if upgrade to .NET Framework 7.3 or .NET 5/6
    internal sealed class EnumAsLowerCaseStringFormatter<T> : IMessagePackFormatter<T>
    {
        private readonly IReadOnlyDictionary<string, T> _nameValueMapping;
        private readonly IReadOnlyDictionary<T, string> _valueNameMapping;

        public EnumAsLowerCaseStringFormatter()
        {
            var fields = typeof(T).GetFields(BindingFlags.DeclaredOnly | BindingFlags.Public | BindingFlags.Static);
            var nameValueMapping = new Dictionary<string, T>();
            var valueNameMapping = new Dictionary<T, string>();

            foreach (FieldInfo enumValueMember in fields)
            {
                string name = enumValueMember.Name.ToLower();
                T value = (T)enumValueMember.GetValue(null);

                nameValueMapping[name] = value;
                valueNameMapping[value] = name;
            }

            _nameValueMapping = nameValueMapping;
            _valueNameMapping = valueNameMapping;
        }

        public void Serialize(ref MessagePackWriter writer, T value, MessagePackSerializerOptions options)
        {
            if (!_valueNameMapping.TryGetValue(value, out string valueString))
            {
                valueString = value.ToString();
            }

            writer.Write(valueString);
        }

        public T Deserialize(ref MessagePackReader reader, MessagePackSerializerOptions options)
        {
            string name = reader.ReadString();

            if (!_nameValueMapping.TryGetValue(name, out T value))
            {
                value = (T)Enum.Parse(typeof(T), name);
            }

            return value;
        }
    }
}
