/**
 * @file    Version.cs
 *
 * @internal
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 * Licensed under the MIT License.
 * Redistributed files must retain the above copyright notice.
 *
 * This file contains the .NET C# SDK build version and the .NET C# SDK API version digit declarations.
 * The build version corresponds to the LMI build that generated the .NET C# SDK.
 *
 * The API version is the version of the interfaces that SDK applications use.
 */

namespace GoPxLSdk
{
    public static class SdkBuildVersionConstants
    {
        /// <summary>
        /// SDK build version major number.
        /// </summary>
        public const uint MAJOR = 8;

        /// <summary>
        /// SDK build version minor number.
        /// </summary>
        public const uint MINOR = 0;

        /// <summary>
        /// SDK build version release number.
        /// </summary>
        public const uint RELEASE = 9;

        /// <summary>
        /// SDK build version build number.
        /// </summary>
        public const uint BUILD = 145;
    }

    public static class SdkApiVersionConstants
    {
        /// <summary>
        /// SDK Api version major number.
        /// </summary>
        public const uint API_MAJOR = 3;

        /// <summary>
        /// SDK Api version minor number.
        /// </summary>
        public const uint API_MINOR = 0;

        /// <summary>
        /// SDK Api version release number.
        /// </summary>
        public const uint API_PATCH = 0;
    }

    public static class SdkVersion
    {
        /// <summary>
        /// Returns the SDK build version as a 32-bit number in the format:
        ///    MMmmRRBB
        /// where:
        ///  MM = major number
        ///  mm = minor number
        ///  RR = release number
        ///  BB = build number
        ///
        /// The SDK build version is the build version that generated the SDK package.
        /// </summary>
        /// <returns>Gocator SDK build version number.</returns>
        public static uint BuildVersionNumber()
        {
            return (SdkBuildVersionConstants.MAJOR & 0xFF) << 24 |
                    (SdkBuildVersionConstants.MINOR & 0xFF) << 16 |
                    (SdkBuildVersionConstants.RELEASE & 0xFF) << 8 |
                    (SdkBuildVersionConstants.BUILD & 0xFF);
        }

        /// <summary>
        /// Returns the SDK build version as a string. Format is:
        ///   MM.mm.RR.BB
        /// where:
        ///  MM = major number
        ///  mm = minor number
        ///  RR = release number
        ///  BB = build number
        ///
        /// The SDK build version is the build version that generated the SDK package.
        /// </summary>
        /// <returns>Gocator SDK build version string.</returns>
        public static string BuildVersion()
        {
            return SdkBuildVersionConstants.MAJOR.ToString() + "." +
                SdkBuildVersionConstants.MINOR.ToString() + "." +
                SdkBuildVersionConstants.RELEASE.ToString() + "." +
                SdkBuildVersionConstants.BUILD.ToString();
        }

        /// <summary>
        /// Returns the SDK API version as string. Format is:
        ///   MM.mm.PN
        /// where:
        ///  MM = major number
        ///  mm = minor number
        ///  PN = patch number
        ///
        /// The SDK API version is the version of the SDK interfaces used to create SDK applications.
        /// </summary>
        /// <returns>Gocator SDK API version string.</returns>
        public static string ApiVersion()
        {
            return "3.0.0";
        }
    }
}
