/*
 * @file            AlignSensor.cs
 * Objective        Connect to a Gocator sensor system to perform alignment.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

using GoPxLSdk;
using Newtonsoft.Json.Linq;
using System;
using System.Net;

namespace AlignSensor
{
    internal class AlignSensorSample
    {
        private const string SENSOR_IP = "192.168.1.10";
        private const int PORT = 3600;
        private const string ENGINE_ID = "g2xxx";
        private const string ALIGN_COMMAND_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/commands/align";
        private const string UNALIGN_COMMAND_PATH = "/scan/engines/g2xxx/scanners/scanner-0/commands/unalign";
        private const string ALIGNMENT_STATE_PATH = "/scan/engines/g2xxx/scanners/scanner-0/alignment";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int REST_COMMAND_TIMEOUT = 3000;

        enum AlignmentStates
        {
            UNALIGNED = 0,
            ALIGNED = 1,
            ALIGNING = 2
        }

        private static int AlignSensor()
        {
            // Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_IP);
            using (var system = new GoSystem(ipAddress, PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                JObject response;
                JObject payload;

                // Ensure job is saved before performing alignment.
                payload = new JObject { ["name"] = "SDK_alignment_sample" };
                system.Client().Call("/jobs/commands/save", payload).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("\nSaved job as " + payload["name"]);

                // Clear alignment.
                Console.WriteLine("\nClearing alignment...");
                system.Client().Call(UNALIGN_COMMAND_PATH).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("Alignment cleared.");

                // Align the sensor.
                Console.WriteLine("\nExecuting alignment...");

                int alignmentState = 0;
                bool aligning = true;
                try
                {
                    system.Client().Call(ALIGN_COMMAND_PATH).CheckResponse(REST_COMMAND_TIMEOUT);
                    
                    // Wait while the sensor is aligning.
                    while (aligning)
                    {
                        System.Threading.Thread.Sleep(1000);
                        response = system.Client().Read(ALIGNMENT_STATE_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload;
                        alignmentState = (int)response["alignState"];
                        if (alignmentState != 2)
                        {
                            aligning = false;
                        }
                    }
                    Console.WriteLine($"Alignment state: {(AlignmentStates)alignmentState}");
                    response = system.Client().Read(ALIGNMENT_STATE_PATH).GetResponse(REST_COMMAND_TIMEOUT).Payload;
                    Console.WriteLine($"Alignment configuration: {response["alignConfig"]}");

                    if (alignmentState == 0)
                    {
                        Console.WriteLine("\nAlignment failed. Check alignment configuration on sensor.");
                    }
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Alignment failed. Check API path, sensor state and availability of scan data.");
                    return ERROR_STATUS;
                }

                system.Disconnect();
                return OK_STATUS;
            }
        }

        static int Main()
        {
            int status = ERROR_STATUS;

            try
            {
                status = AlignSensor();
            }
            catch (GoRequestException ex)
            {
                JToken error = ex.Response.Payload["errors"][0];
                string errorMessage = error != null ? $"Error: {error["status"]}     {error["description"]}" : "";

                Console.WriteLine(errorMessage);
                Console.WriteLine("GoRequestError - Error sending a REST command to " + ex.Response.Path);
            }
            catch (GoChannelException ex)
            {
                Console.WriteLine($"Error: {ex.Message} - Check sensor status and make sure it is connected.");
            }
            catch (Exception ex)
            {
                Console.WriteLine($"Error: {ex.Message}");
            }

            Console.WriteLine("\nPress any key to exit the program...");
            Console.ReadKey(true);
            return status;
        }
    }
}