/*
 * @file            BackupRestore.cs
 * Objective        Connect to a Gocator sensor system, save a backup archive, and restore a backup archive.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

using GoPxLSdk;
using Newtonsoft.Json.Linq;
using System;
using System.Collections;
using System.IO;
using System.Net;

namespace BackupRestore
{
    internal class BackupRestoreSample
    {
        private const string SENSOR_IP = "192.168.1.10";
        private const int PORT = 3600;
        private const string SYSTEM_BACKUP_PATH = "/system/commands/archive";
        private const string SYSTEM_RESTORE_PATH = "/system/commands/restore";
        private const string BACKUP_FILE_PATH = "./backup.gpbak";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        // Ensure timeout value is set properly for your use case.
        const int REST_COMMAND_TIMEOUT = 10_000; // 10 seconds
        const int RECEIVE_DATA_TIMEOUT = 20_000; // Receiving data might require a higher timeout value.

        private static int BackupRestore()
        {
            // Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_IP);
            using (var system = new GoSystem(ipAddress, PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                JObject response;
                JObject payload;

                // Fetch data and download as a file.
                payload = new JObject { ["contents"] = new JArray { "global", "allWorkspaces", "allJobs", "replay" } };
                // To save as a support file, use the following and change the BACKUP_FILE_PATH extension to.gpsup:
                // payload = new JObject { ["contents"] = new JArray { "global", "allWorkspaces", "allJobs", "replay", "support" } };

                // Get data.
                response = system.Client().Call(SYSTEM_BACKUP_PATH, payload).GetResponse(RECEIVE_DATA_TIMEOUT).Payload;
                byte[] data = response["data"].ToObject<byte[]>();
                Console.WriteLine("Data received.");

                // Download.
                try
                {
                    File.WriteAllBytes(BACKUP_FILE_PATH, data);
                    Console.WriteLine("File downloaded to " + Path.GetFullPath(BACKUP_FILE_PATH));
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to download the file to " + Path.GetFullPath(BACKUP_FILE_PATH));
                    return ERROR_STATUS;
                }

                // Upload and restore.
                Console.WriteLine("\nRestoring system...");
                byte[] dataByteArray = File.ReadAllBytes(BACKUP_FILE_PATH);

                var payloadUpdated = new JObject(payload);
                payloadUpdated["data"] = dataByteArray;
                // payloadUpdated.Set("data", data); // to restore using the currently received data.
                try
                {
                    system.Client().Call(SYSTEM_RESTORE_PATH, payloadUpdated).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("System restored using " + Path.GetFullPath(BACKUP_FILE_PATH));
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to restore the system using " + Path.GetFullPath(BACKUP_FILE_PATH));
                    return ERROR_STATUS;
                }

                system.Disconnect();
                return OK_STATUS;
            }
        }

        static int Main()
        {
            int status = ERROR_STATUS;
            try
            {
                status = BackupRestore();
            }
            catch (GoRequestException ex)
            {
                JToken error = ex.Response.Payload["errors"][0];
                string errorMessage = error != null ? $"Error: {error["status"]}     {error["description"]}" : "";

                Console.WriteLine(errorMessage);
                Console.WriteLine("GoRequestError - Error sending a REST command to " + ex.Response.Path);
            }
            catch (GoChannelException ex)
            {
                Console.WriteLine($"Error: {ex.Message} - Check sensor status and make sure it is connected.");
            }
            catch (Exception ex)
            {
                Console.WriteLine($"Error: {ex.Message}");
            }

            Console.WriteLine("\nPress any key to exit the program...");
            Console.ReadKey(true);
            return status;
        }
    }
}