/*
 * @file            ConfigureSensor.cs
 * Objective        Connect to a Gocator sensor system to change sensor configurations.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

using GoPxLSdk;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Runtime.InteropServices;

namespace ConfigureSensor
{
    internal class ConfigureSensorSample
    {
        private const string ENV_CONFIG_PATH = "/environ/ipConfig";
        private const string VISIBLE_SENSORS_PATH = "/scan/visibleSensors/";
        private const string TRIGGER_SOURCE_PATH = "$.parameters.triggerSettings.source";
        private const string EXPOSURE_MODE_PATH = "$.parameters.exposureSettings.exposureMode";
        private const string SINGLE_EXPOSURE_PATH = "$.parameters.exposureSettings.singleExposure";
        private const string MULTI_EXPOSURE_PATH = "$.parameters.exposureSettings.multipleExposures";
        private const string ACTIVE_AREA_X_FOV_PATH = "$.parameters.activeAreaSettings.activeArea.width";
        private const string ENGINE_ID = "g2xxx";
        private const string SENSOR_GROUP_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0";
        private const string SENSOR_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/sensors/sensor-0";
        private const string SENSOR_SERIAL_NUMBER = "95246";
        private const int SINGLE_EXPOSURE_MODE = 0;
        private const int MULTI_EXPOSURE_MODE = 1;
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int REST_COMMAND_TIMEOUT = 3000;

        private static int ConfigureSensor()
        {
            IPAddress ipAddress;
            int port;

            // Connect to a sensor system with a known IP address and port.
            ipAddress = IPAddress.Parse("192.168.1.10");
            port = 3600;

            using (var system = new GoSystem(ipAddress, port))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nApplication connected to control port of sensor system {ipAddress}:{port}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Connection failed. Check if sensor is powered on and connected " +
                                                   "and check you are using the correct IP address and port.");
                    return ERROR_STATUS;
                }

                // Define API endpoints for the connected sensor.
                JToken sensors = system.Client().Read(VISIBLE_SENSORS_PATH).GetResponse().Payload["sensors"];
                if (!sensors.HasValues)
                {
                    Console.WriteLine("Unable to read any visible sensors at path: " + VISIBLE_SENSORS_PATH);
                    return ERROR_STATUS;
                }

                bool endPointFound = false;
                JToken sensor = sensors.First();
                foreach (var sensorIter in sensors)
                {
                    if (sensorIter.SelectToken($".serialNumber").ToString() == SENSOR_SERIAL_NUMBER)
                    {
                        sensor = sensorIter;
                        endPointFound = true;
                        Console.WriteLine("Sensor API endpoint updated to sensor with serial number: " + SENSOR_SERIAL_NUMBER);
                        break;
                    }
                }

                if (!endPointFound)
                {
                    Console.WriteLine("Unable to find sensor API endpoint for the serial number: " + SENSOR_SERIAL_NUMBER);
                    return ERROR_STATUS;
                }

                JObject response;

                // Set parameter path based on engineId
                string triggerSourcePath = TRIGGER_SOURCE_PATH;

                // Configuration 1 - Update Trigger Source:
                Console.WriteLine("Configuring sensor with IP Address: " + sensor.SelectToken($".parameters.ipAddress"));
                response = system.Client().Read(SENSOR_GROUP_PATH).GetResponse().Payload;
                int triggerSource = response.SelectToken(triggerSourcePath).Value<int>();

                var triggerSourceMap = new Dictionary<int, string> { { 0, "Time" }, { 1, "Encoder" }, { 2, "External Input" }, { 3, "Software" } };
                Console.WriteLine($"\nPrevious Trigger Source: {triggerSource} ({triggerSourceMap[triggerSource]})");

                response.SelectToken(triggerSourcePath).Replace(JToken.FromObject(1));
                system.Client().Update(SENSOR_GROUP_PATH, response).CheckResponse(REST_COMMAND_TIMEOUT);
                int updatedTriggerSource = response.SelectToken(triggerSourcePath).Value<int>();
                Console.WriteLine($"Status: {GoStatus.Ok} - Updated Trigger Source: {updatedTriggerSource} ({triggerSourceMap[updatedTriggerSource]})");

                // Configuration 2 - Set/update single exposure:
                Console.WriteLine("SENSOR_PATH: " + SENSOR_PATH + "\n");
                response = system.Client().Read(SENSOR_PATH).GetResponse().Payload;
                Console.WriteLine("\nPrevious Single Exposure value: " + response.SelectToken(SINGLE_EXPOSURE_PATH));

                response.SelectToken(EXPOSURE_MODE_PATH).Replace(JToken.FromObject(SINGLE_EXPOSURE_MODE)); // Set Exposure Mode. 0: Single exposure, 1: Multiple exposures, 2: Dynamic exposure.
                response.SelectToken(SINGLE_EXPOSURE_PATH).Replace(JToken.FromObject(1200.0));
                system.Client().Update(SENSOR_PATH, response).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine($"Status: {GoStatus.Ok} - Updated Single Exposure value: " + response.SelectToken(SINGLE_EXPOSURE_PATH));

                // Configuration 3 - Set/update multiple exposures:
                response = system.Client().Read(SENSOR_PATH).GetResponse().Payload;
                Console.WriteLine("\nPrevious Multiple Exposure value: " + response.SelectToken(MULTI_EXPOSURE_PATH));

                response.SelectToken(EXPOSURE_MODE_PATH).Replace(JToken.FromObject(MULTI_EXPOSURE_MODE));
                int[] multiExposureValues = { 1080, 2010, 5040 };
                for (int i = 0; i < multiExposureValues.Length; i++)
                {
                    response.SelectToken(MULTI_EXPOSURE_PATH + "[" + i + "]").Replace(JToken.FromObject(multiExposureValues[i]));
                }

                system.Client().Update(SENSOR_PATH, response).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine($"Status: {GoStatus.Ok} - Updated Multiple Exposure value: " + response.SelectToken(MULTI_EXPOSURE_PATH));

                // Configuration 4 - Update sensor name:
                response = system.Client().Read(SENSOR_PATH).GetResponse().Payload;
                Console.WriteLine("\nPrevious Display Name: " + response["displayName"]);

                response["displayName"] = "Main-Sensor";

                system.Client().Update(SENSOR_PATH, response).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine($"Status: {GoStatus.Ok} - Updated Display Name: " + response["displayName"]);

                // Configuration 5 - Change the active area: 
                response = system.Client().Read(SENSOR_PATH).GetResponse().Payload;
                double xFov = Convert.ToDouble(response.SelectToken(ACTIVE_AREA_X_FOV_PATH).ToString());
                Console.WriteLine("Previous X FOV: " + xFov + " mm");

                response.SelectToken(ACTIVE_AREA_X_FOV_PATH).Replace(xFov/2);

                system.Client().Update(SENSOR_PATH, response).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine($"Status: {GoStatus.Ok} - Updated Active Area X Fov: " + response[ACTIVE_AREA_X_FOV_PATH]);

                // Configuration 6 - Change network interface configuration:
                response = system.Client().Read(ENV_CONFIG_PATH).GetResponse().Payload;
                Console.WriteLine("\nPrevious network configurations: " + response["interfaces"]);

                // Uncomment to update the current IP address to 192.168.1.11.
                /*
                response["interfaces"][0]["ipAddress"] = "192.168.1.11";
                system.Client().Update(ENV_CONFIG_PATH, response).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine($"Status: {GoStatus.Ok} - Current IP address of the 1st interface: " + response.SelectToken("$.interfaces.0.ipAddress"));
                */

                // Disconnect the sensor system.
                system.Disconnect();
            }
            return OK_STATUS;
        }

        static int Main()
        {
            int status = ERROR_STATUS;

            try
            {
                status = ConfigureSensor();
            }
            catch (GoRequestException ex)
            {
                JToken error = ex.Response.Payload["errors"][0];
                string errorMessage = error != null ? $"Error: {error["status"]} {error["description"]}" : "";

                Console.WriteLine(errorMessage);
                Console.WriteLine("GoRequestError - Error sending a REST command to " + ex.Response.Path);
            }
            catch (GoChannelException ex)
            {
                Console.WriteLine($"Error: {ex.Message} - Check sensor status and make sure it is connected.");
            }
            catch (Exception ex)
            {
                Console.WriteLine($"Error: {ex.Message}");
            }

            Console.WriteLine("\nPress any key to exit the program...");
            Console.ReadKey(true);
            return status;
        }
    }
}
