/*
 * @file            ReceiveAsync.cs
 * Objective        Receive data asynchronously.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 *
 * To receive data over Gocator Data Protocol (GDP), use GoGdpClient::ReceiveDataSync() or GoGdpClient::ReceiveDataAsync().
 *  GoGdpClient::ReceiveDataAsync() works as follows:
 *  1. GoGdpClient::connect() is called to establish connection with the Protocol.
 *  2. GoGdpClient::onDataReceive() starts a receiveThread & receives messages and deserializes them.
 *  3. Individual data messages are received until GoGdpClient::Close() is called.
 *  4. GoGdpClient::ReceiveDataAsync() receives individual data messages and re-assembles to GoDataSet object "dataSet".
 *  5. The new data set is added to the Go::MsgQueue<GoDataSet> dataQueue.
 *  6. In a second dataThread, the data set is removed from dataQueue.
 *  7. Callback function can be used to access dataSet.
 */

using GoPxLSdk;
using GoPxLSdk.GoGdpMsg;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Xml.Linq;

namespace ReceiveAsync
{
    internal class ReceiveAsyncSample
    {
        private const string SENSOR_IP = "192.168.1.10";
        private const int PORT = 3600;
        private const string ENGINE_ID = "g2xxx";
        private const string SCANNER_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0";
        private const string OUTPUT_STAMP_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/outputs/stamp";
        private const string GOCATOR_CONTROL_PATH = "/controls/gocator";
        private const string GOCATOR_ADD_OUTPUT_PATH = "/controls/gocator/outputs/commands/add";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int REST_COMMAND_TIMEOUT = 3_000; // 3 seconds.
        private const int GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT = 3601;

        // Callback function for ReceiveDataAsync()
        private static void OnData(GoDataSet receivedDataSet)
        {
            if (receivedDataSet.Count < 1)
            {
                Console.WriteLine("No data is received.");
                throw new Exception("No data is received.");
            }

            for (int i = 0; i < receivedDataSet.Count; i++)
            {
                switch (receivedDataSet.GdpMsgAt(i).Type)
                {
                    case MessageType.Stamp:
                        {
                            var stampMsg = (GoGdpStamp)receivedDataSet.GdpMsgAt(i);
                            Console.WriteLine("\nFrame index: " + stampMsg.FrameIndex);
                            Console.WriteLine("Timestamp: " + stampMsg.Timestamp);
                            Console.WriteLine("Encoder: " + stampMsg.Encoder);
                            Console.WriteLine("Encoder at Z: " + stampMsg.EncoderAtZ);
                            Console.WriteLine("Status: " + stampMsg.Status);
                            break;
                        }
                    case MessageType.Measurement:
                        {
                            var measurementMsg = (GoGdpMeasurement)receivedDataSet.GdpMsgAt(i);
                            Console.WriteLine("\nMeasurement");
                            Console.WriteLine("Data Source: " + measurementMsg.DataSourceId);
                            Console.WriteLine("Array count: " + measurementMsg.ArrayedCount);
                            Console.WriteLine("Array index: " + measurementMsg.ArrayedIndex);
                            Console.WriteLine("Value: " + measurementMsg.Value);
                            Console.WriteLine("Decision: " + measurementMsg.Decision);
                            Console.WriteLine("Output ID: " + measurementMsg.GdpId);
                            break;
                        }
                }
            }
        }

        static int ReceiveAsync()
        {
            // Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_IP);
            using (var system = new GoSystem(ipAddress, PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                JObject response;
                JObject payload;

                // Set trigger source and frame rate.
                var triggerSourceMap = new Dictionary<int, string> { { 0, "Time" }, { 1, "Encoder" }, { 2, "External Input" }, { 3, "Software" } };
                payload = new JObject { ["parameters"] = "triggerSettings", ["source"] = 0, ["frameRate"] = 0.3 };
                try
                {
                    system.Client().Update(SCANNER_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    response = system.Client().Read(SCANNER_PATH).GetResponse().Payload;
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Check if the API path is valid.");
                    return ERROR_STATUS;
                }

                // Run the sensor.
                if (system.RunningState() == GoSystem.State.Ready)
                {
                    system.Start();
                    Console.WriteLine("\nSystem started.");
                }

                // Enable the Gocator Protocol.
                payload = new JObject { ["enabled"] = true };
                system.Client().Update(GOCATOR_CONTROL_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("\nGocator Protocol enabled.");

                // Adds stamp as a data output.
                response = system.Client().Read(OUTPUT_STAMP_PATH).GetResponse().Payload;
                string stampDataSourceId = response["dataSourceId"].ToString();

                payload = new JObject { ["source"] = stampDataSourceId, ["outputId"] = 0, ["autoShift"] = true };
                system.Client().Call(GOCATOR_ADD_OUTPUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("\nStamp added to Gocator output.");

                // Connect to the Gocator Data Protocol (GDP) server.
                using (var gdpClient = new GoGdpClient())
                {
                    try
                    {
                        gdpClient.Connect(system.Address, GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT);
                        Console.WriteLine("\nGocator Protocol Connected.");
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Error: " + ex.Message + " - Failed to connect to the GDP server.");
                        return ERROR_STATUS;
                    }

                    // Establish an asynchronous data transfer and start receiving data.
                    Console.WriteLine("\nRunning callback function to receive data...");
                    Console.WriteLine("Press ENTER to close the connection and exit.");
                    Action<GoDataSet> func = OnData;
                    try
                    {
                        gdpClient.ReceiveDataAsync(func);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex.Message + " - Failed to receive data. Make sure data is available.");
                        return ERROR_STATUS;
                    }

                    Console.ReadKey(true);
                    gdpClient.Close();
                }

                system.Stop();
                Console.WriteLine("\nSystem stopped.");
                system.Disconnect();
            }

            return OK_STATUS;
        }

        static int Main()
        {
            int status = ReceiveAsync();

            if (status == ERROR_STATUS)
            {
                Console.WriteLine("\nPress any key to exit the program...");
                Console.ReadKey(true);
            }
            return status;
        }
    }
}