/*
 * @file            ReceiveProfile.cs
 * Objective        Connect to the GoPxL Data Protocol and receive profile data.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 *
 * GoGdpClient::ReceiveDataSync() or GoGdpClient::ReceiveDataAsync() can be used to receive data from the GoPxL Data Protocol (GDP) serve:
    1. Pass a timeout value (in microseconds) within which data will be received from the server.
    2. The GDP server sends out individual data messages. The last message is marked with an isLastMsg flag.
    3. ReceiveDataSync() receives individual data messages and reassembles to a data set (GoDataSet).
    4. GoGdpClient::DataSet() can be used to access the data set.
    5. GoGdpMsg::Type() indicates the type of individual messages in the dataset as described in GoPxLSdk::MessageType.
 */

using GoPxLSdk;
using GoPxLSdk.GoGdpMsg;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.Data;
using System.Net;

namespace ReceiveProfile
{
    internal class ReceiveProfileSample
    {
        private const string SENSOR_IP = "192.168.1.10";
        private const int PORT = 3600;
        private const string ENGINE_ID = "g2xxx";
        private const string OUTPUTS_PATH = "/scan/engines/" + ENGINE_ID + "/scanners/scanner-0/outputs";
        private const string OUTPUT_UNIFORM_PROFILE_PATH = OUTPUTS_PATH + "/topUniformProfile";
        private const string GOCATOR_CONTROL_PATH = "/controls/gocator";
        private const string GOCATOR_ADD_OUTPUT_PATH = "/controls/gocator/outputs/commands/add";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int REST_COMMAND_TIMEOUT = 3_000; // 3 seconds.
        private const int RECEIVE_DATA_TIMEOUT = 3_000;

        private const int GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT = 3601;

        // Represents a point in a profile. 
        public struct ProfilePoint
        {
            public double X { get; set; }   // X coordinate - position along laser line (mm).
            public double Z { get; set; }   // Z coordinate - height at a given x position (mm).
            public byte Intensity { get; set; }
        }

        private static int ReceiveProfile()
        {
            // Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_IP);
            using (var system = new GoSystem(ipAddress, PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                JObject response;
                JObject payload;
                ProfilePoint[] profileBuffer;

                // Run the sensor.
                if (system.RunningState() == GoSystem.State.Ready)
                {
                    system.Start();
                    Console.WriteLine("\nSystem started.");
                }

                // Enable the Gocator Protocol.
                payload = new JObject { ["enabled"] = true };
                try
                {
                    system.Client().Update(GOCATOR_CONTROL_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nGocator Protocol enabled.");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Check the timeout value or the API path.");
                    return ERROR_STATUS;
                }

                // Adds profile as a data output.
                payload = new JObject { ["enabled"] = true };
                try
                {
                    response = system.Client().Read(OUTPUT_UNIFORM_PROFILE_PATH).GetResponse().Payload;
                    string profileDataSourceId = response["dataSourceId"].ToString();
                    payload = new JObject { ["source"] = profileDataSourceId, ["outputId"] = 0, ["autoShift"] = true };
                    system.Client().Call(GOCATOR_ADD_OUTPUT_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("Added profile to the Gocator Protocol output.");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to add profile to Gocator Protocol output. Check API path and ensure uniform spacing is enabled.");
                    return ERROR_STATUS;
                }

                // Connect to the Gocator Data Protocol (GDP) server.
                using (var gdpClient = new GoGdpClient())
                {
                    try
                    {
                        gdpClient.Connect(system.Address, GO_PXL_SDK_DEFAULT_GDP_SERVER_PORT);
                        Console.WriteLine("\nGocator Protocol Connected.");
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex.Message + " - Failed to connect to the GDP server.");
                        return ERROR_STATUS;
                    }

                    // Establish a synchronous data transfer.
                    try
                    {
                        gdpClient.ReceiveDataSync(RECEIVE_DATA_TIMEOUT);
                        Console.WriteLine("\nSynchronous data transfer established.");
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("Error: " + ex.Message + " - Failed to receive data via synchronous connection.");
                        return ERROR_STATUS;
                    }

                    // Receive data.
                    GoDataSet receivedDataSet = gdpClient.DataSet;
                    if (receivedDataSet.Count < 1)
                    {
                        Console.WriteLine("\nNo data is received.");
                        return ERROR_STATUS;
                    }
                    Console.WriteLine("\nTotal number of messages: " + receivedDataSet.Count);

                    // Loop through messages in the dataSet.
                    for (int i = 0; i < receivedDataSet.Count; i++)
                    {
                        Console.WriteLine($"\nMessage {i + 1} out of {receivedDataSet.Count}");
                        Console.WriteLine($"Data source: {receivedDataSet.GdpMsgAt(i).DataSourceId}");

                        switch (receivedDataSet.GdpMsgAt(i).Type)
                        {

                            case MessageType.Stamp:
                                {
                                    var stampMsg = (GoGdpStamp)receivedDataSet.GdpMsgAt(i);
                                    Console.WriteLine("\nStamp");
                                    Console.WriteLine("Frame index: " + stampMsg.FrameIndex);
                                    Console.WriteLine("Timestamp: " + stampMsg.Timestamp);
                                    Console.WriteLine("Encoder: " + stampMsg.Encoder);
                                    Console.WriteLine("Encoder at Z: " + stampMsg.EncoderAtZ);
                                    Console.WriteLine("Status: " + stampMsg.Status);
                                    break;
                                }

                            case MessageType.UniformProfile:
                                {
                                    var uniformProfileMsg = (GoGdpProfileUniform)receivedDataSet.GdpMsgAt(i);

                                    Console.WriteLine("\nUniform Profile");
                                    Console.WriteLine("Data Source: " + uniformProfileMsg.DataSourceId);
                                    Console.WriteLine("Array count: " + uniformProfileMsg.ArrayedCount);
                                    Console.WriteLine("Array index: " + uniformProfileMsg.ArrayedIndex);

                                    int validPointCount = 0;
                                    uint profilePointCount = uniformProfileMsg.Width;
                                    profileBuffer = new ProfilePoint[profilePointCount];

                                    /*  The uniform profile data is sent in an array of z values.
                                    *  The x real-world coordinates can be calculated from the index of the array,
                                    *  using the x offset and resolution.
                                    *  The z value must be transformed into real world coordinates before being used.
                                    */

                                    for (int k = 0; k < profilePointCount; k++)
                                    {
                                        short range = uniformProfileMsg.Ranges[k];

                                        profileBuffer[k] = new ProfilePoint();
                                        profileBuffer[k].X = uniformProfileMsg.Offset.X + uniformProfileMsg.Resolution.X * k;

                                        if (range != short.MinValue)
                                        {
                                            // Calculates the Z position in mm from the given data.
                                            // Z = Z offset + Z resolution * z
                                            profileBuffer[k].Z = uniformProfileMsg.Offset.Z + uniformProfileMsg.Resolution.Z * range;
                                            validPointCount++;
                                        }
                                        else
                                        {
                                            profileBuffer[k].Z = short.MinValue;
                                        }
                                    }
                                    Console.WriteLine("Profile points count: " + profilePointCount);
                                    Console.WriteLine("Valid points count: " + validPointCount);
                                    break;
                                }

                            // Get raw profile data.
                            case MessageType.ProfilePointCloud:
                                {
                                    var pointCloudProfileMsg = (GoGdpProfilePointCloud)receivedDataSet.GdpMsgAt(i);

                                    Console.WriteLine("\nUniform Profile");
                                    Console.WriteLine("Data Source: " + pointCloudProfileMsg.DataSourceId);
                                    Console.WriteLine("Array count: " + pointCloudProfileMsg.ArrayedCount);
                                    Console.WriteLine("Array index: " + pointCloudProfileMsg.ArrayedIndex);

                                    int validPointCount = 0;
                                    uint profilePointCount = pointCloudProfileMsg.Width;
                                    profileBuffer = new ProfilePoint[profilePointCount];

                                    /*  The non-uniform profile data is sent in an array of x and z values.
                                     *  The x and z values must be transformed into real world coordinates,
                                     *  using the x and z offset and resolution values.
                                     */

                                    for (int k = 0; k < profilePointCount; k++)
                                    {
                                        GoPoint16s point = pointCloudProfileMsg.Ranges[k];

                                        if (point.X != short.MinValue)
                                        {
                                            // Calculates the X and Z position in mm from the given data.
                                            // X = X offset + X resolution * x
                                            // Z = Z offset + Z resolution * z
                                            profileBuffer[k].X = pointCloudProfileMsg.Offset.X + pointCloudProfileMsg.Resolution.X * point.X;
                                            profileBuffer[k].Z = pointCloudProfileMsg.Offset.Z + pointCloudProfileMsg.Resolution.Z * point.Y;
                                            validPointCount++;
                                        }
                                        else
                                        {
                                            profileBuffer[k].X = short.MinValue;
                                            profileBuffer[k].Z = short.MinValue;
                                        }
                                    }
                                    Console.WriteLine("Profile points count: " + profilePointCount);
                                    Console.WriteLine("Valid points count: " + validPointCount);
                                    break;
                                }

                            default:
                                Console.WriteLine("\nNo Profile or Stamp received.\n");
                                break;
                        }
                    }

                    gdpClient.Close();
                }

                Console.WriteLine("\nPress any key to continue...");
                system.Stop();
                Console.WriteLine("\nSystem stopped.");
                system.Disconnect();
            }

            return OK_STATUS;
        }

        static int Main()
        {
            int status = ReceiveProfile();
            Console.WriteLine("\nPress any key to exit the program...");
            Console.ReadKey(true);
            return status;
        }
    }
}