/*
 * @file            SaveJob.cs
 * Objective        Connect to a Gocator sensor system, list saved jobs, save and load a new job.
 * 
 * GoPxLSdk Sample
 * Copyright (C) 2022-2023 by LMI Technologies Inc.
 *
 * Licensed under the MIT License.
 * Redistributions of files must retain the above copyright notice.
 */

using GoPxLSdk;
using Newtonsoft.Json.Linq;
using System;
using System.Collections;
using System.IO;
using System.Linq;
using System.Net;
using System.Xml.Linq;

namespace SaveJob
{
    internal class SaveJobSample
    {
        private const string SENSOR_IP = "192.168.1.10";
        private const int PORT = 3600;
        private const string VISIBLE_SENSORS_PATH = "/scan/visibleSensors/";
        private const string JOBS_PATH = "/jobs";
        private const string JOB_FILES_PATH = "/jobs/files";
        private const string SAVE_JOB_PATH = "/jobs/commands/save";
        private const string LOAD_JOB_PATH = "/jobs/commands/load";
        private const string RENAME_JOB_PATH = "/jobs/commands/rename";
        private const string LOCAL_JOB_FILE_PATH = "./sample_job.gpjob";
        private const int ERROR_STATUS = -1;
        private const int OK_STATUS = 0;
        private const int REST_COMMAND_TIMEOUT = 3_000; // 3 seconds.
        private const int RECEIVE_DATA_TIMEOUT = 20_000;

        private static int SaveJob()
        {
            // Connect to a sensor system.
            IPAddress ipAddress;
            ipAddress = IPAddress.Parse(SENSOR_IP);
            using (var system = new GoSystem(ipAddress, PORT))
            {
                try
                {
                    system.Connect();
                    Console.WriteLine($"\nConnected to {ipAddress}:{PORT}");
                }
                catch (Exception ex)
                {
                    Console.WriteLine(
                        ex.Message
                            + " - Connection failed. Check if sensor is powered on and connected "
                            + "also check if you are using the correct IP address and port number."
                    );
                    return ERROR_STATUS;
                }

                JObject response;
                JObject payload;

                // Get the list of jobs.
                var expandLevel = new JObject { ["expandLevel"] = 1 };
                response = system.Client().Read(JOB_FILES_PATH, args: expandLevel).GetResponse().Payload;
                if (response.SelectToken("_embedded.item") != null)
                {
                    JToken jobs = response.SelectToken("_embedded.item");
                    Console.WriteLine("\nList of saved jobs: ");
                    foreach (var it in jobs)
                    {
                        Console.WriteLine(it.SelectToken("jobName"));
                    }
                }
                else
                {
                    Console.WriteLine("\nCurrently GoPxL does not contain any saved job.");
                }

                // Save job.
                payload = new JObject { ["name"] = "SDK-demo" };
                system.Client().Call(SAVE_JOB_PATH, payload).CheckResponse(REST_COMMAND_TIMEOUT);
                Console.WriteLine("\nSaved the current job as " + payload["name"]);

                // Load job.
                response = system.Client().Read(JOBS_PATH, args: expandLevel).GetResponse().Payload;
                Console.WriteLine("\nLoaded job: " + response["loadedJob"].Value<string>());

                // Rename job.
                var renameJobContent = new JObject { ["sourceName"] = "SDK-demo", ["destName"] = "SDK-demo-job" };
                try
                {
                    system.Client().Call(RENAME_JOB_PATH, renameJobContent).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nJob renamed to " + renameJobContent["destName"]);
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to rename the job. Make sure the job names are correct and the destination name does not already exist.");
                }

                // Download job.
                var READ_DATA_PATH = "/jobs/files/SDK-demo-job/data";
                response = system.Client().Read(READ_DATA_PATH, payload).GetResponse(RECEIVE_DATA_TIMEOUT).Payload;
                byte[] jobData = response["content"].ToObject<byte[]>();
                Console.WriteLine("\nJob data received.");
                try
                {
                    File.WriteAllBytes(LOCAL_JOB_FILE_PATH, jobData);
                    Console.WriteLine("\nJob downloaded to " + Path.GetFullPath(LOCAL_JOB_FILE_PATH));
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to download the job to " + Path.GetFullPath(LOCAL_JOB_FILE_PATH));
                    return ERROR_STATUS;
                }

                // Upload job.
                jobData = File.ReadAllBytes(LOCAL_JOB_FILE_PATH);
                var localJobContent = new JObject { ["fromLive"] = false, ["name"] = "SDK-local-job-demo", ["content"] = jobData };
                try
                {
                    system.Client().Create(JOB_FILES_PATH, localJobContent).CheckResponse(REST_COMMAND_TIMEOUT);
                    Console.WriteLine("\nUpload job as " + localJobContent["name"] + " from " + Path.GetFullPath(LOCAL_JOB_FILE_PATH));
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message + " - Failed to upload job from " + Path.GetFullPath(LOCAL_JOB_FILE_PATH));
                    Console.WriteLine("Make sure the job name does not already exist.");
                }
                
                system.Disconnect();
                return OK_STATUS;
            }
        }

        static int Main()
        {
            int status = SaveJob();
            Console.WriteLine("\nPress any key to exit the program...");
            Console.ReadKey(true);
            return status;
        }
    }
}